import pkg_resources
import stevedore
from e3.platform_db import get_knowledge_base


def test_knownledge_base():
    """Test loading a e3.platform_db extension."""
    # Create a new entry point referencing AmberCPUSupport and load it.
    mydb_ep = pkg_resources.EntryPoint.parse_group(
        "e3.platform_db",
        ["mydb = e3.platform_db:AmberCPUSupport"],
        dist=pkg_resources.get_distribution("e3-core"),
    )["mydb"]
    mydb_ep.load()

    # Inject it in the ExtensionManager entry point cache

    old_cache = stevedore.ExtensionManager.ENTRY_POINT_CACHE
    try:
        stevedore.ExtensionManager.ENTRY_POINT_CACHE["e3.platform_db"].append(mydb_ep)

        db = get_knowledge_base()
        # Check that the new amber CPU has been inserted
        for cpu_name in ("amber23", "amber25"):
            assert cpu_name not in db.cpu_info

        # Force a reload of the platform_db knowledge base
        get_knowledge_base.cache_clear()
        db = get_knowledge_base()

        # Check that the new amber CPU has been inserted
        for cpu_name in ("amber23", "amber25"):
            assert db.cpu_info[cpu_name] == {"endian": "little", "bits": 32}
    finally:
        # restore entry point cache
        stevedore.ExtensionManager.ENTRY_POINT_CACHE = old_cache



from __future__ import annotations

import abc
import logging
from itertools import chain
from typing import TYPE_CHECKING

from e3.anod.status import ReturnValue
from e3.job import EmptyJob
from e3.job.scheduler import DEFAULT_JOB_MAX_DURATION, Scheduler

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, FrozenSet, Optional, Set
    from e3.collection.dag import DAG
    from e3.fingerprint import Fingerprint
    from e3.job import Job, ProcessJob


class Walk:
    """An abstract class scheduling and executing a DAG of actions.

    :ivar actions: DAG of actions to perform.
    :vartype actions: DAG
    :ivar prev_fingerprints: A dict of e3.fingerprint.Fingerprint objects,
          indexed by the corresponding job ID. This dictionary contains
          the former fingerprints a given job or None if there was no
          such fingerprint.
          (with the job corresponding to a given entry in the DAG of
          actions).
    :vartype prev_fingerprints: Dict[str, Optional[Fingerprint]]
    :ivar new_fingerprints: A dict of e3.fingerprint.Fingerprint objects,
        indexed by the corresponding job ID. This dictionary contains
        the fingerprints we compute each time we create a new job
        (with the job corresponding to a given entry in the DAG of
        actions).
    :vartype new_fingerprints: Dict[str, Optional[Fingerprint]]
    :ivar job_status: A dictionary of job status (ReturnValue), indexed by
        job unique IDs.
    :vartype job_status: Dict[str, ReturnValue]
    :ivar scheduler: The scheduler used to schedule and execute all
        the actions.
    :vartype scheduler: e3.job.scheduler.Scheduler
    """

    def __init__(self, actions: DAG):
        """Object initializer.

        :param actions: DAG of actions to perform.
        """
        self.actions = actions
        self.prev_fingerprints: Dict[str, Optional[Fingerprint]] = {}
        self.new_fingerprints: Dict[str, Optional[Fingerprint]] = {}
        self.job_status: Dict[str, ReturnValue] = {}
        self.set_scheduling_params()
        self.failure_source: Dict[str, Set[str]] = {}

        self.scheduler = Scheduler(
            job_provider=self.get_job,
            collect=self.collect,  # type: ignore
            queues=self.queues,
            tokens=self.tokens,
            job_timeout=self.job_timeout,
        )

        self.scheduler.run(self.actions)

    def set_scheduling_params(self) -> None:
        """Set the parameters used when creating the scheduler.

        This method is expected to set the following attributes:
            - self.queues: Same as parameter of the same name
                in e3.job.scheduler.Scheduler.__init__.
            - self.tokens: Likewise.
            - self.job_timeout: Likewise.

        This method provides a default setup where the scheduler has
        1 token and 1 queue, and where the job's maximum duration is
        DEFAULT_JOB_MAX_DURATION. Child classes requiring different
        scheduling parameters should override this method.
        """
        self.queues = None
        self.tokens = 1
        self.job_timeout = DEFAULT_JOB_MAX_DURATION

    def compute_fingerprint(
        self, uid: str, data: Any, is_prediction: bool = False
    ) -> Optional[Fingerprint]:
        """Compute the given action's Fingerprint.

        This method is expected to return a Fingerprint corresponding
        to the state of the system should the given action be executed
        and succesful. It can also return None if keeping track of
        the result of past actions is not necessary.

        This implementation always returns None. Child classes requiring
        a different behavior may override this method.

        :param uid: A unique Job ID.
        :param data: Data associated to the job.
        :param is_prediction: If True this is an attempt to compute the
            fingerprint before launching the job. In that case if the
            function returns None then the job will always be launched.
            When False, this is the computation done after running the
            job (that will be the final fingerprint saved for future
            comparison)
        """
        return None

    def save_fingerprint(self, uid: str, fingerprint: Optional[Fingerprint]) -> None:
        """Save the given fingerprint.

        For systems that require fingerprint persistence, this method
        is expected to save the fingerprint somewhere -- typically
        inside a file. Passing None as the fingerprint causes
        the fingerprint to be deleted instead.

        This implementation does nothing. Child classes taking advantage
        of fingerprint support should override this method to save
        the fingerprint at the location of their choice and using
        the format that suits them.

        :param uid: A unique Job ID.
        :param fingerprint: The fingerprint corresponding to the given
            job, or None, if the fingerprint should be deleted instead
            of saved.
        """
        pass

    def load_previous_fingerprint(self, uid: str) -> Optional[Fingerprint]:
        """Get the fingerprint from the given action's previous execution.

        This method is expected to have the following behavior:
            - If the given action has already previously been executed
              and its fingerprint saved (see method "save_fingerprint"
              above), then load and return it;
            - Otherwise, return None.

        This implementation always returns None, providing a behavior
        where re-executing a given action always results in the
        corresponding job being executed. Child classes requiring
        a different behavior may override this method.

        :param uid: A unique Job ID.
        """
        return None

    def should_execute_action(
        self,
        uid: str,
        previous_fingerprint: Optional[Fingerprint],
        new_fingerprint: Optional[Fingerprint],
    ) -> bool:
        """Return True if the given action should be performed.

        The purpose of this function is to provide a way to determine,
        based on the previous fingerprint, and the new one, whether
        the user of this class wants us to launch the action or not.

        The default implementation implements the following strategy:
            - when fingerprints are not in use, always execution
              the given action;
            - when fingerprints are in use, execute the given action
              if the fingerprint has changed.
        However, child classes may want to override this method
        to implement alternative strategies.

        :param uid: A unique Job ID.
        :param previous_fingerprint: The fingerprint from the previous
            from the action's previous run.  None if the action has not
            been previously executed.
        :param new_fingerprint: The fingerprint from the previous
            from the action's previous run.  None if the action has not
            been previously executed.
        """
        if previous_fingerprint is None or new_fingerprint is None:
            return True
        for pred_uid in self.actions.get_predecessors(uid):
            if self.new_fingerprints[str(pred_uid)] is None:
                # One of the predecessors has no fingerprint, so
                # this node's new_fingerprint cannot tell us whether
                # this dependency has changed or not. We therefore
                # need to run this action.
                return True
        return previous_fingerprint != new_fingerprint

    def create_skipped_job(
        self,
        uid: str,
        data: Any,
        predecessors: FrozenSet[str],
        reason: Optional[str],
        notify_end: Callable[[str], None],
        status: ReturnValue = ReturnValue.failure,
    ) -> Job:
        """Return a failed job.

        This method always returns an EmptyJob. Deriving classes may
        override this method if they need something more specific.

        :param uid: A unique Job ID.
        :param data: Data associated to the job to create.
        :param predecessors: A list of predecessor jobs, or None.
        :param reason: If not None, the reason for creating a failed job.
        :notify_end: Same as the notify_end parameter in Job.__init__.
        """
        return EmptyJob(uid, data, notify_end, status=status)

    @abc.abstractmethod
    def create_job(
        self,
        uid: str,
        data: Any,
        predecessors: FrozenSet[str],
        notify_end: Callable[[str], None],
    ) -> ProcessJob:
        """Create a ProcessJob.

        :param uid: A unique Job ID.
        :param data: Data associated to the job to create.
        :param predecessors: A list of predecessor jobs, or None.
        :notify_end: Same as the notify_end parameter in Job.__init__.
        """
        pass  # all: no cover

    @abc.abstractmethod
    def request_requeue(self, job: ProcessJob) -> bool:
        """Requeue the given job.

        Return True if the job has been requeued, False otherwise.

        :param job: The job to requeue.
        """
        pass  # all: no cover

    def get_job(
        self,
        uid: str,
        data: Any,
        predecessors: FrozenSet[str],
        notify_end: Callable[[str], None],
    ) -> Job:
        """Return a Job.

        Same as self.create_job except that this function first checks
        whether any of the predecessors might have failed, in which case
        the failed job (creating using the create_skipped_job method)
        is returned.
        """
        # Get the latest fingerprint
        self.prev_fingerprints[uid] = self.load_previous_fingerprint(uid)

        # And reset the fingerprint on disk
        self.save_fingerprint(uid, None)

        self.new_fingerprints[uid] = self.compute_fingerprint(
            uid, data, is_prediction=True
        )

        # Check our predecessors. If any of them failed, then return
        # a failed job.
        failed_predecessors = [
            k
            for k in predecessors
            if self.job_status[k]
            not in (
                ReturnValue.success,
                ReturnValue.skip,
                ReturnValue.force_skip,
                ReturnValue.unchanged,
            )
        ]
        if failed_predecessors:
            force_fail = "Event failed because of prerequisite failure:\n"
            force_fail += "\n".join(
                ["  " + str(self.actions[k]) for k in failed_predecessors]
            )

            # Compute the set of job that originate that failure
            self.failure_source[uid] = set(
                chain.from_iterable(
                    [self.failure_source.get(k, [k]) for k in failed_predecessors]
                )
            )
            force_fail += "\n\nOrigin(s) of failure:\n"
            force_fail += "\n".join(
                ["  " + str(self.actions[k]) for k in self.failure_source[uid]]
            )

            return self.create_skipped_job(
                uid,
                data,
                predecessors,
                force_fail,
                notify_end,
                status=ReturnValue.force_fail,
            )

        if self.should_execute_action(
            uid, self.prev_fingerprints[uid], self.new_fingerprints[uid]
        ):
            return self.create_job(uid, data, predecessors, notify_end)
        else:
            return self.create_skipped_job(
                uid, data, predecessors, "skipped", notify_end, status=ReturnValue.skip
            )

    def collect(self, job: ProcessJob) -> bool:
        """Collect all the results from the given job.

        :param job: The job whose results we need to collect.

        :return: True if the job is requeued, False otherwise
        """
        # Only save the fingerprint if the job went as expected (either
        # success or skipped). Since we already removed the previous
        # fingerprint when we created the job, not saving the fingerprint
        # ensures that we try that action again next time (as opposed
        # to skipping it).
        if job.status in (
            ReturnValue.success,
            ReturnValue.force_skip,
            ReturnValue.skip,
            ReturnValue.unchanged,
        ):
            self.new_fingerprints[job.uid] = self.compute_fingerprint(job.uid, job.data)
            self.save_fingerprint(job.uid, self.new_fingerprints[job.uid])

        self.job_status[job.uid] = ReturnValue(job.status)

        if job.should_skip:
            if job.status not in (ReturnValue.force_fail, ReturnValue.force_skip):
                logging.info(
                    "[%-10s %-9s %4ds] %s",
                    job.queue_name,
                    self.job_status[job.uid].name,
                    0,
                    job.data,
                )
            return False

        logging.info(
            "[%-10s %-9s %4ds] %s",
            job.queue_name,
            job.status.name,
            int(job.timing_info.duration),
            job.data,
        )

        requeued = False
        if self.job_status[job.uid] == ReturnValue.notready:
            requeued = self.request_requeue(job)

        return requeued



from __future__ import annotations

import stevedore


def load_cache(name="file-cache", configuration=None):
    plugin = stevedore.DriverManager(namespace="e3.store.cache.backend", name=name)
    return plugin.driver(configuration)



# Check that spec_config is in globals is need just because of
# style checking. This is not required in practic
if "spec_config" not in globals():
    spec_config = None

if spec_config is not None:
    PROLOG_WAS_EXECUTED = True  # type: ignore



from __future__ import annotations

import hashlib
import json
import os
import tempfile
from contextlib import closing
from typing import TYPE_CHECKING

import e3.log
from e3.anod.status import ReturnValue
from e3.os.process import PIPE
from e3.fs import VCS_IGNORE_LIST, get_filetree_state, rm, sync_tree
from e3.vcs.git import GitError, GitRepository
from e3.vcs.svn import SVNError, SVNRepository

if TYPE_CHECKING:
    from typing import Callable, List, Optional, Tuple

logger = e3.log.getLogger("e3.anod.checkout")


class CheckoutManager:
    """Helper class to manage checkouts done by Anod tools.

    When a checkout manager is used in working_dir directory for a repository
    called name. The following structure will be found on disk after an update

        working_dir/name/               The repository checkout
                   /name_checkout.json  A json containing some metadata: name,
                                        url, old_commit, new_commit, revision
                   /name_changelog.json A json containing the list of commit
                                        between two call to update. If this is
                                        the inital checkout or there are no
                                        changes then the file will not be
                                        created
    """

    def __init__(self, name: str, working_dir: str, compute_changelog: bool = True):
        """Initialize CheckoutManager instance.

        :param name: a symbolic name for that checkout
        :param working_dir: working directory in which checkouts are
            performed. Note that the checkout will be done in the
            ```name``` subdirectory.
        :param compute_changelog: if True compute a changelog of changes
            done since last update
        """
        self.name = name
        self.compute_changelog = compute_changelog
        self.working_dir = os.path.abspath(os.path.join(working_dir, self.name))
        self.metadata_file = self.working_dir + "_checkout.json"
        self.changelog_file = self.working_dir + "_changelog.json"

    def update(self, vcs: str, url: str, revision: Optional[str] = None) -> ReturnValue:
        """Update content of the working directory.

        :param vcs: vcd kind
        :param url: repository url
        :param revision: revision
        """
        # Reset changelog file
        if os.path.isfile(self.changelog_file):
            rm(self.changelog_file)

        update: Callable[
            [str, Optional[str]], Tuple[ReturnValue, Optional[str], Optional[str]]
        ]
        if vcs == "git":
            update = self.update_git
        elif vcs == "svn":
            update = self.update_svn
        elif vcs == "external":
            update = self.update_external
        else:
            logger.error("Invalid repository type: %s" % vcs)
            return ReturnValue.failure

        result, old_commit, new_commit = update(url=url, revision=revision)

        with open(self.metadata_file, "w") as fd:
            json.dump(
                {
                    "name": self.name,
                    "url": url,
                    "old_commit": old_commit,
                    "new_commit": new_commit,
                    "revision": revision,
                },
                fd,
            )
        return result

    def update_external(
        self, url: str, revision: Optional[str]
    ) -> Tuple[ReturnValue, str, str]:
        """Update working dir using a local directory.

        :param url: path to the repository
        :param revision: ignored
        """
        if os.path.isdir(self.working_dir):
            old_commit = get_filetree_state(self.working_dir)
        else:
            old_commit = ""
        ignore_list: List[str] = []

        if os.path.isdir(os.path.join(url, ".git")):
            # It seems that this is a git repository. Get the list of files to
            # ignore
            try:
                g = GitRepository(working_tree=url)
                ignore_list_lines = g.git_cmd(
                    [
                        "ls-files",
                        "-o",
                        "--ignored",
                        "--exclude-standard",
                        "--directory",
                    ],
                    output=PIPE,
                ).out
                ignore_list = [
                    "/%s" % f.strip().rstrip("/")
                    for f in ignore_list_lines.splitlines()
                ]
                logger.debug("Ignore in external: %s", ignore_list)
            except Exception:
                # don't crash on exception
                pass

        sync_tree(
            url,
            self.working_dir,
            preserve_timestamps=False,
            delete_ignore=True,
            ignore=list(VCS_IGNORE_LIST) + ignore_list,
        )

        new_commit = get_filetree_state(self.working_dir)
        if new_commit == old_commit:
            return ReturnValue.unchanged, old_commit, new_commit
        else:
            return ReturnValue.success, old_commit, new_commit

    def update_git(
        self, url: str, revision: Optional[str]
    ) -> Tuple[ReturnValue, Optional[str], Optional[str]]:
        """Update working dir using a Git repository.

        :param url: git repository url
        :param revision: git revision
        """
        # For git repositories revision cannot be None
        if revision is None:
            return ReturnValue.failure, None, None

        g = GitRepository(working_tree=self.working_dir)
        g.log_stream = e3.log.default_output_stream
        old_commit, new_commit = None, None

        # Create a remote for which name is bind to a an url
        # This ensure that when the url does not change, git will not
        # redownload all the objects on each objects (and thus avoid
        # disk space leaks).
        remote_name = hashlib.sha256(url.encode("utf-8")).hexdigest()
        g.init()

        # Do the remote addition manually as in that context we can ignore
        # safely any error returned by this command.
        try:
            remote_list = g.git_cmd(["remote"], output=PIPE).out.splitlines()
            if remote_name not in remote_list:
                g.git_cmd(["remote", "add", remote_name, url])
        except Exception:
            # Ignore exception as it probably means that remote already exist
            # In case of real error the failure will be detected later.
            pass

        try:
            old_commit = g.rev_parse()

            # Using fetch + checkout ensure caching is effective
            g.git_cmd(["fetch", "-f", remote_name, "%s:refs/e3-checkout" % revision])
            g.checkout("refs/e3-checkout", force=True)
            new_commit = g.rev_parse()

            # Verify that there is no local change
            p = g.git_cmd(["status", "--porcelain"], output=PIPE)
            if p.out:
                logger.error(
                    "Repository %s is locally modified, saving " "diff in stash\n%s",
                    self.name,
                    p.out,
                )
                g.git_cmd(["stash", "save", "-u"])

            if old_commit == new_commit:
                result = ReturnValue.unchanged
            elif self.compute_changelog:
                # Fetch the change log and dump it into the changelog file
                with closing(tempfile.NamedTemporaryFile(mode="w", delete=False)) as fd:
                    g.write_log(fd, rev_range=f"{old_commit}..{new_commit}")
                    tmp_filename = fd.name
                try:
                    with open(tmp_filename) as fd:
                        commits = list(g.parse_log(fd, max_diff_size=1024))
                finally:
                    rm(tmp_filename)

                with open(self.changelog_file, "w") as fd:
                    json.dump(commits, fd)
                # We have removed local changes or updated the git repository
                result = ReturnValue.success
            else:
                result = ReturnValue.success

        except GitError:
            logger.exception("Error during git update %s" % self.name)
            result = ReturnValue.failure
        return result, old_commit, new_commit

    def update_svn(
        self, url: str, revision: Optional[str]
    ) -> Tuple[ReturnValue, Optional[str], Optional[str]]:
        """Update working dir using a SVN repository.

        :param url: git repository url
        :param revision: git revision
        """
        working_copy = SVNRepository(working_copy=self.working_dir)
        working_copy.log_stream = e3.log.default_output_stream
        old_commit, new_commit = None, None
        result = ReturnValue.success

        if os.path.isdir(self.working_dir):  # a possible checkout exists
            try:
                old_commit = working_copy.current_revision
            except SVNError:
                logger.error(
                    "Unable to get SVN informations form the %s " "working dir",
                    self.name,
                )
        try:
            # Remove local change and update the working copy
            local_change_detected = working_copy.update(
                url=url, revision=revision, force_and_clean=True
            )
        except SVNError:  # impossible to update, potential local changes
            logger.error("Impossible to update the working copy of %s", self.name)
            result = ReturnValue.failure
        else:
            new_commit = working_copy.current_revision
            if local_change_detected:
                logger.error(
                    "Repository %s was locally modified," " clean done.", self.name
                )
            if old_commit == new_commit:
                result = ReturnValue.unchanged
            else:
                result = ReturnValue.success
        return result, old_commit, new_commit



import e3.anod.action as action
import e3.anod.spec
import e3.collection.dag
import e3.env


def test_initall():
    """Very simple test to check that all objects can be initialized."""
    root = action.Root()
    assert root.uid == "root"
    assert str(root) == "root node"

    class MySource:
        name = "my_source"

    class Spec(e3.anod.spec.Anod):
        uid = "my_source_uid"
        name = "my_source_spec"

    get_source = action.GetSource(builder=MySource())
    assert get_source.uid == "source_get.my_source"
    assert str(get_source) == "get source my_source"

    download_source = action.DownloadSource(builder=MySource())
    assert download_source.uid == "download.my_source"
    assert str(download_source) == "download source my_source"

    install_source = action.InstallSource(
        uid="install.my_source",
        spec=Spec(qualifier="", kind="source"),
        source=MySource(),
    )
    assert str(install_source) == "install source my_source"

    create_sources = action.CreateSources(
        anod_instance=Spec(qualifier="", kind="source")
    )
    assert str(create_sources) == "create all sources for my_source_spec.anod"

    create_source = action.CreateSource(
        anod_instance=Spec(qualifier="", kind="source"), source_name="my_source"
    )
    assert str(create_source) == "create source my_source"

    checkout = action.Checkout(repo_name="e3-core", repo_data={})
    assert str(checkout) == "checkout e3-core"

    build_spec = Spec(qualifier="", kind="build")
    build_spec.name = "my_spec"
    build_spec.env = e3.env.Env()

    build = action.Build(anod_instance=build_spec)
    assert str(build).startswith("build my_spec for ")

    test_spec = Spec(qualifier="", kind="test")
    test_spec.name = "my_spec"
    test_spec.env = e3.env.Env()

    test = action.Test(anod_instance=test_spec)
    assert str(test).startswith("test my_spec for ")

    install_spec = Spec(qualifier="", kind="build")
    install_spec.name = "my_spec"
    install_spec.env = e3.env.Env()

    install = action.Install(anod_instance=build_spec)
    assert str(install).startswith("build my_spec for ")

    download_spec = Spec(qualifier="", kind="build")
    download_spec.name = "my_spec"
    download_spec.env = e3.env.Env()

    download = action.DownloadBinary(data=build_spec)
    assert str(download).startswith("download binary of ")

    upload_spec = Spec(qualifier="", kind="build")
    upload_spec.name = "my_spec"
    upload_spec.env = e3.env.Env()

    upload_bin = action.UploadBinaryComponent(data=build_spec)
    assert str(upload_bin).startswith("upload binary package of")

    upload_source = action.UploadSourceComponent(data=build_spec)
    assert str(upload_source).startswith("upload source metadata of ")

    a_decision = action.CreateSourceOrDownload(
        root=root, left=create_source, right=download_source
    )

    assert a_decision.get_decision() is None

    a_decision.set_decision(action.Decision.RIGHT, "plan_line.txt:1")
    assert a_decision.get_decision() == "download.my_source"

    a_decision.set_decision(action.Decision.BOTH, "plan_line.txt:2")
    assert a_decision.get_decision() is None

    boi_decision = action.BuildOrDownload(root=install, left=build, right=download)
    assert boi_decision.get_decision() is None
    boi_decision.set_decision(action.Decision.LEFT, "plan_line.txt:3")
    assert boi_decision.get_decision() == build.uid


def test_trigger():
    """Test Decision triggers."""

    class Spec(e3.anod.spec.Anod):
        uid = "my_source_uid"
        name = "my_source_spec"

    dag = e3.collection.dag.DAG()

    root = action.Root()

    build_spec = Spec(qualifier="", kind="build")
    build_spec.name = "my_spec"
    build_spec.env = e3.env.Env()
    build = action.Build(anod_instance=build_spec)

    build_spec2 = Spec(qualifier="", kind="build")
    build_spec2.name = "my_spec2"
    build_spec2.env = e3.env.Env()
    build2 = action.Build(anod_instance=build_spec2)

    download2 = action.DownloadBinary(build_spec2)

    install_spec2 = Spec(qualifier="", kind="install")
    install_spec2.name = "my_spec2"
    install_spec2.env = e3.env.Env()
    install2 = action.Install(install_spec2)

    dag.add_vertex(root)
    dag.add_vertex(install2)

    # Add a trigger on 'build'
    decision = action.BuildOrDownload(root=install2, left=build2, right=download2)
    decision.add_trigger(build, action.Decision.LEFT, "my_plan_line")

    decision.apply_triggers(dag)

    assert decision.get_expected_decision() is None
    # 'build' not in dag --> expected decision is None

    dag.add_vertex(build.uid, build)
    decision.apply_triggers(dag)
    assert decision.get_expected_decision() == build2.uid

    # Now change the decision, applying the trigger again should show the
    # conflict
    decision.expected_choice = action.Decision.RIGHT
    decision.apply_triggers(dag)
    assert decision.get_expected_decision() is None  # Decision.BOTH

    # Set back to the expected choice and ask for another one
    decision.expected_choice = build2.uid
    decision.set_decision(action.Decision.RIGHT, "my_other_plan_line")
    assert decision.get_decision() is None  # Decision.BOTH

    # Also change the expected choice to RIGHT
    decision.expected_choice = action.Decision.RIGHT
    assert decision.get_expected_decision() == download2.uid



from __future__ import annotations

from enum import Enum


class ReturnValue(Enum):
    """Return values for anod commands."""

    success = 0
    failure = 1
    missing = 2
    notready = 75
    force_skip = 122
    force_fail = 123
    unknown = 124
    skip = 125
    unchanged = 126



import sys

from e3.anod.status import ReturnValue
from e3.job import EmptyJob, ProcessJob


class TestJob:
    def test_run_empty_job(self):
        """Try running an empty job...

        In theory, trying to run an EmptyJob does not make sense,
        since the job's should_skip is True. But, in practice,
        verify that nothing bad happens if someone tries it.

        Also, this allows us to cover this part of the code,
        since the e3.job.scheduler.Scheduler class does not
        do it for us (for the reason specified above).
        """
        job = EmptyJob("1", None, print, ReturnValue.success)
        assert job.should_skip is True
        assert job.status == ReturnValue.success

        job.start(1)
        assert job.should_skip is True
        assert job.status == ReturnValue.success

    def test_process_job_status_before_run(self):
        """Check the status of a ProcessJob before running it.

        Normally, we don't expect anyone to try this, but in case
        it happens, this unit-test makes sure nothing bad happens.
        """

        class EmptyProcessJob(ProcessJob):
            def cmdline(self):
                return [sys.executable, "-c", "pass"]

        job = EmptyProcessJob("1", None, print)
        assert job.status is ReturnValue.notready

    def test_invalid_job_status(self, caplog):
        """Verify that when the return code is invalid we return a failure."""

        class InvalidProcessJob(ProcessJob):
            cmdline = [sys.executable, "-c", "import sys; sys.exit(6)"]

        job = InvalidProcessJob("myuid", {}, None)
        job.run()
        assert job.status is ReturnValue.failure
        assert "job myuid returned an unknown status 6" in caplog.text

    def test_spawn_issue(self):
        """Verify that status is set to failure when the spawn fails."""

        class SpawnIssueProcessJob(ProcessJob):
            @property
            def cmdline(self):
                # This will be called by self.run() and simulate an error
                # when spawning the job
                raise OSError("spawn issue")

        job = SpawnIssueProcessJob("myuid", {}, None)
        job.run()
        assert job.status is ReturnValue.failure



"""Provides function to detect platform specific information."""
from __future__ import annotations

import os
import re
import sys
from collections import namedtuple
from platform import uname as platform_uname
from typing import TYPE_CHECKING

import e3.log
from e3.platform_db import get_knowledge_base


if TYPE_CHECKING:
    from typing import Optional, Tuple, Union

KNOWLEDGE_BASE = get_knowledge_base()

UNKNOWN = "unknown"

Uname = namedtuple(
    "Uname", ["system", "node", "release", "version", "machine", "processor"]
)


class SystemInfo:
    """Gather info about the system.

    :cvar network_ifs: dictionary addressed by network interface name for which
        each value is the result of netifaces.ifaddresses function on the given
        interface
    :cvar uname: instance of Uname namedtuple containing the result of
        ``uname`` system call.
    :cvar core_number: integer containing the number of processor cores on the
        machine
    :cvar nis_domain: host nis domain
    """

    network_ifs = None
    uname = None
    core_number = 1
    nis_domain = None
    ld_info = None

    # Cache for SystemInfo methods
    _platform: Optional[str] = None

    # _os_version is a tuple: os version, kernel version
    _os_version: Optional[Tuple[str, str]] = None

    # When _is_virtual is True it means the host is a VM
    _is_virtual: Optional[bool] = None

    # _hostname is a tuble: hostname, domain. Joining with a dot hostname and domain
    # represent the FQDN
    _hostname: Optional[Tuple[str, str]] = None

    @classmethod
    def reset_cache(cls) -> None:
        """Reset SystemInfo cache."""
        cls.network_ifs = None
        cls.uname = None
        cls.ld_info = None
        cls.core_number = 1
        cls._is_virtual = None
        cls._platform = None
        cls._os_version = None
        cls._hostname = None

    @classmethod
    def fetch_system_data(cls) -> None:
        """Fetch info from the host system.

        The function should be the only one that use system calls or programs
        to fetch information from the current system. Overriding this method
        should be enough for example for testing purposes as all the other
        methods use information retrieved in this function.

        The function should set all the class attributes described at the
        beginning of this class.
        """
        # Compute result of uname
        cls.uname = Uname(*platform_uname())

        # Fetch linux distribution info on linux OS
        if cls.uname.system == "Linux":  # linux-only
            import ld

            cls.ld_info = {
                "name": ld.name(),
                "major_version": ld.major_version(),
                "version": ld.version(),
            }

        # Fetch network interfaces
        try:
            from netifaces import interfaces, ifaddresses, address_families

            # use string for address families instead of integers which are
            # system dependents
            cls.network_ifs = {
                itf: {address_families[k]: v for k, v in ifaddresses(itf).items()}
                for itf in interfaces()
            }
        except Exception:  # defensive code
            e3.log.debug("cannot get network info", exc_info=True)
            cls.network_ifs = None

        # Fetch core numbers. Note that the methods does not work
        # on AIX platform but we usually override manually that
        # setting anyway.
        cls.core_number = 1
        try:
            import multiprocessing

            cls.core_number = multiprocessing.cpu_count()
        except Exception:  # defensive code
            e3.log.debug("multiprocessing error", exc_info=True)
            try:
                import psutil

                cls.core_number = psutil.cpu_count()
            except Exception:
                e3.log.debug("psutil error", exc_info=True)

        cls.nis_domain = UNKNOWN

        if sys.platform != "win32":  # windows: no cover
            try:
                import nis
            except ImportError:  # defensive code
                e3.log.debug("cannot import nis", exc_info=True)
                nis = None  # type: ignore

            if nis is not None:
                try:
                    cls.nis_domain = nis.get_default_domain()
                    if not cls.nis_domain:  # defensive code
                        cls.nis_domain = UNKNOWN
                except nis.error:  # defensive code
                    e3.log.debug("nis error", exc_info=True)

    @classmethod
    def platform(cls) -> str:
        """Guess platform name.

        Internal function that guess base on uname system call the
        current platform

        :return: the platform name
        """
        if cls._platform is not None:
            return cls._platform

        if cls.uname is None:
            cls.fetch_system_data()

        if TYPE_CHECKING:
            assert cls.uname is not None

        result = [
            p
            for p, v in KNOWLEDGE_BASE.host_guess.items()
            if cls.uname.system == v["os"]
            and (
                v["cpu"] is None
                or re.match(v["cpu"], cls.uname.machine)
                or re.match(v["cpu"], cls.uname.processor)
            )
        ]

        if result:
            cls._platform = result[0]
        else:  # defensive code
            cls._platform = UNKNOWN

        return cls._platform

    @classmethod
    def os_version(cls) -> Tuple[str, str]:
        """Compute OS version information.

        :return: a tuple containing os version and kernel version
        """
        if cls._os_version is not None:
            return cls._os_version

        if cls.uname is None:
            cls.fetch_system_data()

        if TYPE_CHECKING:
            assert cls.uname is not None
            assert cls.ld_info is not None

        version = UNKNOWN
        kernel_version = UNKNOWN
        system = cls.uname.system

        if system == "Darwin":  # darwin-only
            version = cls.uname.release
        elif system == "FreeBSD":  # bsd-only
            version = re.sub("-.*", "", cls.uname.release)
        elif system == "Linux":  # linux-only
            kernel_version = cls.uname.release
            name = cls.ld_info["name"].lower()
            if "redhat" in name or "red hat" in name:  # os-specific
                name = "rhES"
                version_number = cls.ld_info["major_version"]
            elif "suse" in name or "sles" in name:  # os-specific
                name = "suse"
                version_number = cls.ld_info["major_version"]
            elif "debian" in name:  # os-specific
                name = "debian"
                version_number = cls.ld_info["major_version"]
            else:  # os-specific
                version_number = cls.ld_info["version"]
            version = name + version_number
        elif system == "AIX":  # aix-only
            version = cls.uname.version + "." + cls.uname.release
        elif system == "SunOS":  # solaris-only
            version = "2" + cls.uname.release[1:]
        elif system == "Windows":  # windows-only
            version = cls.uname.release.replace("Server", "")
            kernel_version = cls.uname.version
            # Compute real underlying OS version. Starting with Windows 8.1
            # (6.3), the win32 function that returns the version may return
            # the wrong version depending on the application manifest. So
            # python will always return Windows 8 in that case.
            import ctypes

            class WinOSVersion(ctypes.Structure):
                _fields_ = [
                    ("dwOSVersionInfoSize", ctypes.c_ulong),
                    ("dwMajorVersion", ctypes.c_ulong),
                    ("dwMinorVersion", ctypes.c_ulong),
                    ("dwBuildNumber", ctypes.c_ulong),
                    ("dwPlatformId", ctypes.c_ulong),
                    ("szCSDVersion", ctypes.c_wchar * 128),
                    ("wServicePackMajor", ctypes.c_ushort),
                    ("wServicePackMinor", ctypes.c_ushort),
                    ("wSuiteMask", ctypes.c_ushort),
                    ("wProductType", ctypes.c_byte),
                    ("wReserved", ctypes.c_byte),
                ]

            def get_os_version() -> Union[
                Tuple[None, None, None], Tuple[float, int, bool]
            ]:
                """Return the real Windows kernel version.

                On recent version, the kernel version returned by the
                GetVersionEx Win32 function depends on the way the application
                has been compiled. Using RtlGetVersion kernel function ensure
                that the right version is returned.

                :return: the version as a table
                    (major.minor, build number, is_server)
                """
                os_version = WinOSVersion()
                os_version.dwOSVersionInfoSize = ctypes.sizeof(os_version)
                retcode = ctypes.windll.Ntdll.RtlGetVersion(  # type: ignore
                    ctypes.byref(os_version)
                )
                if retcode != 0:
                    return (None, None, None)

                return (
                    float(f"{os_version.dwMajorVersion}.{os_version.dwMinorVersion}"),
                    os_version.dwBuildNumber,
                    os_version.wProductType != 1,
                )

            effective_version, build_number, is_server = get_os_version()

            if effective_version is None or effective_version <= 6.2:
                if version == "Vista" and "64" in cls.uname.machine:  # os-specific
                    version = "Vista64"
            else:
                if effective_version == 6.3:
                    if is_server:
                        version = "2012R2"
                    else:
                        version = "8.1"
                elif effective_version == 10.0:
                    if TYPE_CHECKING:
                        assert build_number is not None
                    if is_server:
                        if build_number < 17763:
                            version = "2016"
                        else:
                            version = "2019"
                    else:
                        version = "10"

        cls._os_version = (version, kernel_version)
        return version, kernel_version

    @classmethod
    def is_virtual(cls) -> bool:
        """Check if current machine is virtual or not.

        :return: True if the machine is a virtual machine (Solaris zone,
            VmWare)
        """
        if cls._is_virtual is not None:
            return cls._is_virtual

        if cls.uname is None:
            cls.fetch_system_data()

        if TYPE_CHECKING:
            assert cls.uname is not None

        result = False

        if (
            cls.uname.system == "SunOS" and cls.uname.version == "Generic_Virtual"
        ):  # solaris-only
            result = True
        else:
            if cls.network_ifs is not None:
                for interface in list(cls.network_ifs.values()):
                    for family in ("AF_LINK", "AF_PACKET"):
                        if family in interface:
                            for el in interface[family]:
                                addr = el["addr"].lower()
                                if addr.startswith("00:0c:29") or addr.startswith(
                                    "00:50:56"
                                ):
                                    result = True
                                    break
                        if result:
                            break
                    if result:
                        break
        cls._is_virtual = result
        return result

    @classmethod
    def hostname(cls) -> Tuple[str, str]:
        """Get hostname and associated domain.

        :return: a tuple (hostname, domain)
        """
        if cls._hostname is not None:
            return cls._hostname

        if cls.uname is None:
            cls.fetch_system_data()

        if TYPE_CHECKING:
            assert cls.uname is not None

        # This is host so we can find the machine name using uname fields
        tmp = cls.uname.node.lower().split(".", 1)
        hostname = tmp[0]
        if len(tmp) > 1:
            domain = tmp[1]
        else:
            domain = cls.nis_domain

        # Hostname can be overriden by E3_HOSTNAME env variable
        hostname = os.environ.get("E3_HOSTNAME", hostname)
        cls._hostname = (hostname, domain)
        return cls._hostname


class CPU(namedtuple("CPU", ["name", "bits", "endian", "cores"])):
    """Object representing a CPU.

    CPU attributes are:

    - name: [str] the CPU name
    - bits: number of bits for the cpu or 'unknown'
    - endian: big, little, or unknown
    - cores: number of cores available
    """

    __slots__ = ()

    def as_dict(self):
        return self._asdict()

    @classmethod
    def get(
        cls, name: str, endian: Optional[str] = None, compute_cores: bool = False
    ) -> CPU:
        """Initialize CPU instance.

        :param name: cpu name
        :param endian: if not None override endianness default settings
        :param compute_cores: if True compute the number of cores
        """
        assert name in KNOWLEDGE_BASE.cpu_info, "invalid cpu name"
        bits = KNOWLEDGE_BASE.cpu_info[name]["bits"]
        cores = 1

        if endian is None:
            endian = KNOWLEDGE_BASE.cpu_info[name]["endian"]
        if compute_cores:
            cores = SystemInfo.core_number

        return CPU(name, bits, endian, cores)


class OS(
    namedtuple(
        "OS",
        [
            "name",
            "version",
            "kernel_version",
            "exeext",
            "dllext",
            "is_bareboard",
            "mode",
        ],
    )
):
    """Object representing an OS.

    Attributes are:

    - name: the OS name
    - version: the OS version
    - kernel_version: the exact version of the kernel
    - exeext: the default executable extension (e.g. .exe on Windows)
    - dllext: the default shared library extension (e.g. .dll on Windows)
    - is_bareboard: whether the system has an OS or not.
    """

    __slots__ = ()

    def as_dict(self):
        return self._asdict()

    @classmethod
    def get(
        cls,
        name: str,
        is_host: bool = False,
        version: str = UNKNOWN,
        mode: str = UNKNOWN,
    ) -> OS:
        """Initialize OS instance.

        :param name: os name
        :param is_host: if True the OS instance is for the host system
        :param version: os version
        :param mode: os mode
        """
        is_bareboard = KNOWLEDGE_BASE.os_info[name]["is_bareboard"]
        if name.startswith("vxworks") and mode == "rtp":
            exeext = ".vxe"
        else:
            exeext = KNOWLEDGE_BASE.os_info[name]["exeext"]

        dllext = KNOWLEDGE_BASE.os_info[name]["dllext"]

        # If version is not given by the user guess it or set it to the
        # default (cross case)
        if version == UNKNOWN:
            if is_host:
                version, kernel_version = SystemInfo.os_version()
            else:
                version = KNOWLEDGE_BASE.os_info[name]["version"]
                kernel_version = UNKNOWN
        else:
            kernel_version = UNKNOWN
        return OS(name, version, kernel_version, exeext, dllext, is_bareboard, mode)



import e3.os.platform


def test_system_info():
    """Compute SystemInfo and test the cache system."""
    s = e3.os.platform.SystemInfo
    s.reset_cache()
    platform_name = s.platform()
    os_version = s.os_version()
    assert s._platform == platform_name
    s._platform = "foo"
    assert s.platform() == "foo"
    s.reset_cache()
    assert s._platform is None
    assert s._os_version is None
    assert s.os_version() == os_version


def test_is_virtual():
    """Check detection of virtual machines."""
    s = e3.os.platform.SystemInfo
    is_virtual = s.is_virtual()
    s.reset_cache()
    assert s.is_virtual() == is_virtual

    # Second call should use the cache
    assert s.is_virtual() == is_virtual

    # Detect virtual machine (fake the network here)
    s.network_ifs = {"eth0": {"AF_LINK": [{"addr": "00:0c:29:dc:40:00"}]}}
    s._is_virtual = None
    assert s.is_virtual() is True


def test_hostname():
    """Test SystemInfo.hostname."""
    s = e3.os.platform.SystemInfo
    try:
        s.reset_cache()
        hostname = s.hostname()
        s.reset_cache()
        assert s.hostname() == hostname
        real_uname = s.uname
        s.reset_cache()
        s.uname = e3.os.platform.Uname(
            system=real_uname.system,
            node="foo.example.net",
            release=real_uname.release,
            version=real_uname.version,
            machine=real_uname.machine,
            processor=real_uname.processor,
        )
        assert s.hostname() == ("foo", "example.net")
    finally:
        #  reset hostname() to the default value
        s.reset_cache()


def test_cpu():
    """Simple test for CPU."""
    cpu = e3.os.platform.CPU.get("x86_64")
    assert cpu.as_dict()["bits"] == 64


def test_os():
    """Simple test for OS."""
    os = e3.os.platform.OS.get("linux")
    assert os.as_dict()["dllext"] == ".so"

    e3.os.platform.KNOWLEDGE_BASE.os_info[
        "vxworks"
    ] = e3.os.platform.KNOWLEDGE_BASE.os_info["linux"]

    os = e3.os.platform.OS.get("vxworks", mode="rtp")
    assert os.as_dict()["exeext"] == ".vxe"
    os = e3.os.platform.OS.get("vxworks", version="6.8", mode="rtp")
    assert os.as_dict()["kernel_version"] == e3.os.platform.UNKNOWN



import cgi
import logging
import os
import threading
import time

import requests_toolbelt.multipart
from e3.net.http import HTTPSession

try:
    from http.server import BaseHTTPRequestHandler, HTTPServer
except ImportError:
    from http.server import BaseHTTPRequestHandler, HTTPServer


class RetryHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if hasattr(self.server, "tries"):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"OK")
            self.wfile.close()
            self.server.tries += 1
        else:
            self.server.tries = 1


class RetryAbortHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if hasattr(self.server, "tries"):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.send_header("Content-Lenght", "4500")
            self.end_headers()
            self.wfile.write(b"OK")
            time.sleep(1.0)
        else:
            self.server.tries = 1


class ContentDispoHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.send_header("content-disposition", 'attachment;filename="dummy.txt"')
        self.end_headers()
        self.wfile.write(b"Dummy!")
        self.wfile.close()


class ServerErrorHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if not hasattr(self.server, "calls"):
            self.server.calls = 1
        else:
            self.server.calls += 1
        self.send_response(500)
        self.end_headers()
        self.wfile.close()

    def do_POST(self):
        self.send_response(500)
        self.end_headers()


class MultiPartPostHandler(BaseHTTPRequestHandler):
    def do_POST(self):
        if "Content-Type" not in self.headers:
            self.send_response(200)
            self.end_headers()
            return

        content = self.rfile.read(int(self.headers["Content-Length"]))
        decoder = requests_toolbelt.multipart.MultipartDecoder(
            content, self.headers["Content-Type"]
        )

        logging.debug("POST received")
        self.server.test_payloads = {}
        for part in decoder.parts:
            logging.debug(list(part.headers.keys()))
            # With python 3.x requests_toolbelt returns bytes
            _, value = cgi.parse_header(
                part.headers[b"Content-Disposition"].decode("utf-8")
            )
            self.server.test_payloads[value["name"]] = part.text

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(b"OK")
        logging.debug("POST finish")


def run_server(handler, func):
    server = HTTPServer(("localhost", 0), handler)
    try:
        thread = threading.Thread(target=server.serve_forever)
        thread.daemon = True
        thread.start()
        base_url = "http://%s:%s/" % server.server_address
        func(server, base_url)
    finally:
        server.shutdown()
        server.server_close()


class TestHTTP:
    def test_session(self):
        """Create a session in a context."""
        with HTTPSession():
            pass

    def test_retry(self):
        def func(server, base_url):
            with HTTPSession() as session:
                session.set_max_retries(base_url, connect=5)
                result = session.download_file(base_url + "dummy", dest=".")
                with open(result, "rb") as fd:
                    content = fd.read()
                assert content == b"OK"
                assert server.tries == 2

        run_server(RetryHandler, func)

    def test_content_dispo(self):
        def func(server, base_url):
            with HTTPSession() as session:
                result = session.download_file(base_url + "dummy", dest=".")
                with open(result, "rb") as fd:
                    content = fd.read()
                assert content == b"Dummy!"
                assert os.path.basename(result) == "dummy.txt"

        run_server(ContentDispoHandler, func)

    def test_content_validation(self):
        def validate(path):
            return False

        def func(server, base_url):
            with HTTPSession() as session:
                result = session.download_file(
                    base_url + "dummy", dest=".", validate=validate
                )
                assert result is None

        run_server(ContentDispoHandler, func)

    def test_error(self):
        def func(server, base_url):
            with HTTPSession() as session:
                result = session.download_file(base_url + "dummy", dest=".")
                assert result is None

        run_server(ServerErrorHandler, func)

    def test_fallback(self):
        def func(server, base_url):
            def inner_func(server2, base_url2):
                logging.info(f"servers: {base_url}, {base_url2}")
                with HTTPSession(base_urls=[base_url, base_url2]) as session:
                    session.set_max_retries(connect=4)
                    result = session.download_file(base_url + "dummy", dest=".")
                    assert result is not None
                    assert server.calls == 1
                    assert server2.tries == 2

            run_server(RetryHandler, inner_func)

        run_server(ServerErrorHandler, func)

    def test_content_abort(self):
        def func(server, base_url):
            def inner_func(server2, base_url2):
                logging.info(f"servers: {base_url}, {base_url2}")
                with HTTPSession(base_urls=[base_url, base_url2]) as session:
                    session.DEFAULT_TIMEOUT = (10.0, 0.2)
                    session.set_max_retries(connect=4)
                    result = session.download_file("dummy", dest=".")
                    assert result is None
                    assert server.calls == 1
                    assert server2.tries == 1

            run_server(RetryAbortHandler, inner_func)

        run_server(ServerErrorHandler, func)

    def test_post_stream_data(self):
        def outter_func(nok_server, nok_url):
            def func(server, url):
                with HTTPSession(base_urls=[nok_url, url]) as session:
                    session.DEFAULT_TIMEOUT = (3.0, 3.0)
                    with open("./data.txt", "wb") as fd:
                        fd.write(b"Hello!")
                    with open("./data.txt", "rb") as fd:
                        session.request(
                            "POST",
                            "dummy",
                            data_streams={
                                "data.txt": fd,
                                "str_metadata": "string",
                                "metadata": {"key1": "val1"},
                            },
                        )
                assert server.test_payloads["metadata"] == '{"key1": "val1"}'
                assert server.test_payloads["data.txt"] == "Hello!"
                assert server.test_payloads["str_metadata"] == "string"

            run_server(MultiPartPostHandler, func)

        run_server(ServerErrorHandler, outter_func)



from __future__ import annotations

import json
import logging

import e3.log
from e3.event import EventHandler


class LoggingHandler(EventHandler):
    def __init__(self, logger_name="", level=logging.DEBUG):
        self.logger_name = logger_name
        self.level = level
        self.log = e3.log.getLogger(logger_name)

    def send_event(self, event):
        d = event.as_dict()
        self.log.log(self.level, json.dumps(d, indent=2))

    def decode_config(self, config_str):
        logger_name, level = config_str.split(",")
        level = int(level)
        return {"logger_name": logger_name, "level": level}

    def encode_config(self):
        return f"{self.logger_name},{self.level}"



import smtplib
from email.utils import make_msgid

import e3.net.smtp
from e3.error import E3Error

import mock
import pytest


def test_sendmail():
    from_addr = "e3@example.net"
    to_addresses = ["info@example.net", "info@example.com"]
    msg_as_string = "test mail content"

    with mock.patch("smtplib.SMTP_SSL") as mock_smtp:
        e3.net.smtp.sendmail(from_addr, to_addresses, msg_as_string, ["smtp.localhost"])

        smtp_mock = mock_smtp.return_value
        assert smtp_mock.sendmail.called
        assert smtp_mock.sendmail.call_count == 1
        smtp_mock.sendmail.assert_called_once_with(
            from_addr, to_addresses, msg_as_string
        )


def test_sendmail_onerror(caplog):
    from_addr = "e3@example.net"
    to_addresses = ["info@example.net", "info@example.com"]
    msg_as_string = "test mail content"
    msg_size_exceed = "A" * 1200
    result = e3.net.smtp.sendmail(
        from_addr, to_addresses, msg_size_exceed, ["smtp.localhost"], max_size=1 / 1024
    )
    assert result is False
    assert "message file too big" in caplog.text

    with mock.patch("smtplib.SMTP_SSL") as mock_smtp:
        mock_smtp.side_effect = smtplib.SMTPException()
        result = e3.net.smtp.sendmail(
            from_addr, to_addresses, msg_as_string, ["smtp.localhost"]
        )
        assert result is False

    with mock.patch("smtplib.SMTP_SSL") as mock_smtp:
        smtp_mock = mock_smtp.return_value
        smtp_mock.sendmail.side_effect = smtplib.SMTPException()
        result = e3.net.smtp.sendmail(
            from_addr, to_addresses, msg_as_string, ["smtp.localhost"]
        )
        assert result is False

    with mock.patch("smtplib.SMTP_SSL") as mock_smtp:
        smtp_mock = mock_smtp.return_value
        smtp_mock.sendmail.return_value = {}

        mid = make_msgid()
        result = e3.net.smtp.sendmail(
            from_addr, to_addresses, msg_as_string, ["smtp.localhost"], message_id=mid
        )
        assert result is True
        assert "Message-ID: %s sent successfully" % mid in caplog.text
        assert "smtp quit" in caplog.text

    with mock.patch("smtplib.SMTP_SSL") as mock_smtp:
        smtp_mock = mock_smtp.return_value
        smtp_mock.sendmail.return_value = {}

        def error_on_quit():
            raise smtplib.SMTPException("error on quit ignored")

        smtp_mock.quit = error_on_quit
        mid = make_msgid()
        result = e3.net.smtp.sendmail(
            from_addr, to_addresses, msg_as_string, ["smtp.localhost"], message_id=mid
        )
        assert result is True
        assert "Message-ID: %s sent successfully" % mid in caplog.text


def test_send_message():
    from_addr = "e3@example.net"
    to_addresses = ["info@example.net", "info@example.com"]
    msg_content = "test mail content"
    msg_subject = "test mail subject"

    with mock.patch("smtplib.SMTP_SSL") as mock_smtp:
        smtp_mock = mock_smtp.return_value
        smtp_mock.sendmail.return_value = {}
        e3.net.smtp.send_message(
            from_addr, to_addresses, msg_subject, msg_content, ["smtp.localhost"]
        )

        assert smtp_mock.sendmail.called
        assert smtp_mock.sendmail.call_count == 1

        smtp_mock.sendmail.return_value = {"error": 2}

        with pytest.raises(E3Error):
            e3.net.smtp.send_message(
                from_addr, to_addresses, msg_subject, msg_content, ["smtp.localhost"]
            )



from __future__ import annotations

import fnmatch
import re
from difflib import unified_diff
from typing import TYPE_CHECKING

import e3.error
import e3.log
import e3.os.process

if TYPE_CHECKING:
    from typing import Callable, List, Optional, Tuple, Union

logger = e3.log.getLogger("diff")


class DiffError(e3.error.E3Error):
    pass


def diff(
    a: Union[str, List[str]],
    b: Union[str, List[str]],
    ignore: Optional[str] = None,
    item1name: str = "expected",
    item2name: str = "output",
    ignore_white_chars: bool = True,
) -> str:
    """Compute diff between two files or list of strings.

    :param a: a filename or a list of strings
    :param b: a filename or a list of strings
    :param ignore: all lines matching this pattern in both files are
        ignored during comparison. If set to None, all lines are considered.
    :param item1name: name to display for a in the diff
    :param item2name: name to display for b in the diff
    :param ignore_white_chars: if True (default) then empty lines,
        trailing and leading white chars on each line are ignored

    :return: A diff string. If the string is equal to '' it means that there
        is no difference
    """
    contents: List[List[str]] = [[], []]

    # Read first item
    if isinstance(a, list):
        contents[0] = a
    else:
        try:
            with open(a) as f:
                contents[0] = f.readlines()
        except OSError:
            contents[0] = []

    # Do same thing for the second one
    if isinstance(b, list):
        contents[1] = b
    else:
        try:
            with open(b) as f:
                contents[1] = f.readlines()
        except OSError:
            contents[1] = []

    # Filter empty lines in both items and ignore white chars at beginning
    # and ending of lines
    for k in (0, 1):
        if ignore_white_chars:
            contents[k] = [
                "%s\n" % line.strip() for line in contents[k] if line.strip()
            ]
        else:
            # Even if white spaces are not ignored we should ensure at
            # that we don't depend on platform specific newline
            contents[k] = ["%s\n" % line.rstrip("\r\n") for line in contents[k]]

        # If we have a filter apply it now
        if ignore is not None:
            contents[k] = [
                line for line in contents[k] if re.search(ignore, line) is None
            ]

    return "".join(unified_diff(contents[0], contents[1], item1name, item2name, n=1))


def patch(
    patch_file: str,
    working_dir: str,
    discarded_files: Optional[Union[List[str], Callable[[str], bool]]] = None,
    filtered_patch: Optional[str] = None,
) -> None:
    """Apply a patch, ignoring changes in files matching discarded_files.

    :param patch_file: the file containing the patch to apply
    :param working_dir: the directory where to apply the patch
    :param discarded_files: list of files or glob patterns (or function taking
        a filename and returning a boolean - True if the file should be
        discarded)
    :param filtered_patch: name of the filtered patch. By default append
        '.filtered' to the patch_file name
    """

    def apply_patch(fname: str) -> None:
        """Run the patch command.

        :raise DiffError: when the patch command fails
        """
        cmd = ["patch", "-p0", "-f"]
        p = e3.os.process.Run(cmd, cwd=working_dir, input=fname)
        if p.status != 0:
            raise DiffError(
                origin="patch",
                message="running %s < %s in %s failed with %s"
                % (" ".join(cmd), fname, working_dir, p.out),
            )
        logger.debug(p.out)

    if discarded_files is None:
        apply_patch(patch_file)
        return

    if filtered_patch is None:
        filtered_patch = patch_file + ".filtered"

    files_to_patch = 0

    with open(patch_file, newline="") as f, open(
        filtered_patch, "w", newline=""
    ) as fdout:

        # Two line headers that mark beginning of patches
        header1: Union[Tuple, Tuple[str, str]] = ()
        header2: Union[Tuple, Tuple[str, str]] = ()
        header2_regexp = None
        # whether the current patch line should discarded
        discard = False

        def write_line(line: str) -> None:
            """Write line in filtered patch.

            :param l: the line to write
            """
            if not discard:
                fdout.write(line)

        for line in f:
            if not header1:
                # Check if we have a potential start of a 2 lines patch header
                m = re.search(r"^[\*-]{3} ([^ \t\n]+)", line)
                if m is None:
                    write_line(line)
                else:
                    header1 = (line, m.group(1))
                    if line[0] == "-":
                        header2_regexp = r"^\+{3} ([^ \n\t]+)"
                    else:
                        header2_regexp = r"^-{3} ([^ \n\t]+)"
            elif not header2:
                # Check if line next to a header first line confirm that that
                # this is the start of a new patch
                assert header2_regexp is not None
                m = re.search(header2_regexp, line)
                if m is None:
                    write_line(header1[0])
                    header1 = ()
                    write_line(line)
                else:
                    header2 = (line, m.group(1))
            else:
                # This is the start of patch. Decide whether to discard it or
                # not
                discard = False
                path_list = [fn for fn in (header1[1], header2[1]) if fn != "/dev/null"]
                if callable(discarded_files):
                    for fn in path_list:
                        if discarded_files(fn):
                            logger.debug(f"patch {patch_file} discarding {fn}")
                            discard = True
                            break
                else:
                    for pattern in discarded_files:
                        for fn in path_list:
                            if fnmatch.fnmatch(fn, pattern):
                                logger.debug(f"patch {patch_file} discarding {fn}")
                                discard = True
                                break
                        if discard:
                            break
                if not discard:
                    files_to_patch += 1
                    write_line(header1[0])
                    write_line(header2[0])
                    write_line(line)
                header1 = ()
                header2 = ()
        # Dangling lines
        if header1:
            write_line(header1[0])
        if header2:  # defensive code
            write_line(header2[0])

    if files_to_patch:
        apply_patch(filtered_patch)
    else:
        logger.debug("All %s content has been discarded", patch_file)



import e3.text


def test_bytes_as_str():
    """Test bytes_as_str function."""
    result = e3.text.bytes_as_str(b"\x00\xff")
    assert result == "\\x00\\xff"
    result = e3.text.bytes_as_str(b"echo\n")
    assert result == "echo\n"



from __future__ import annotations

from typing import TYPE_CHECKING

import e3.error

if TYPE_CHECKING:
    from typing import Optional
    from e3.os.process import Run

    assert Run  # Make pyflakes happy?


class AnodError(e3.error.E3Error):
    """Base Anod error."""

    pass


class SpecError(AnodError):
    """Invalid specification file."""

    pass


class SandBoxError(AnodError):
    """Invalid sandbox or wrong sandbox configuration."""

    pass


class ShellError(AnodError):
    """Error returned by a process spawned by Anod."""

    def __init__(
        self,
        message: str,
        origin: Optional[str] = None,
        process: Optional["Run"] = None,
    ):
        super().__init__(message, origin)
        self.process = process



from __future__ import annotations

import os
import sys

import yaml
from pkg_resources import get_distribution
from typing import TYPE_CHECKING

import e3.log
import e3.os.process
from e3.anod.buildspace import BuildSpace
from e3.anod.error import SandBoxError
from e3.env import Env
from e3.fs import mkdir, rm
from e3.os.fs import chmod

logger = e3.log.getLogger("sandbox")


if TYPE_CHECKING:
    from typing import Any, Dict, Optional


class SandBox:
    def __init__(self) -> None:
        self.__root_dir: Optional[str] = None
        self.build_id: Optional[str] = None
        self.build_date: Optional[str] = None
        self.build_version: Optional[str] = None

        # Required directories for a sandbox
        self.dirs = (
            "meta",
            "bin",
            "tmp",
            os.path.join("tmp", "cache"),
            "src",
            "log",
            "vcs",
            "patch",
        )

        self.meta_dir: Optional[str] = None
        self.tmp_dir: Optional[str] = None
        self.tmp_cache_dir: Optional[str] = None
        self.src_dir: Optional[str] = None
        self.log_dir: Optional[str] = None
        self.vcs_dir: Optional[str] = None
        self.patch_dir: Optional[str] = None
        self.bin_dir: Optional[str] = None
        self.__specs_dir: Optional[str] = None
        self.is_alternate_specs_dir = False

        # Contains the loaded version of user.yaml if present
        self.user_config: Optional[Dict[str, Any]] = None

        self.default_env = {
            "LANG": "C",
            "LC_ALL": "C",
            "LD_LIBRARY_PATH": "",
            "LIBRARY_PATH": "",
            "DYLD_LIBRARY_PATH": "",
            "DYLD_FALLBACK_LIBRARY_PATH": "",
            "PKG_CONFIG_PATH": "",
            "ADA_PROJECT_PATH": "",
            "GPR_PROJECT_PATH": "",
            "CPATH": "",
            "C_INCLUDE_PATH": "",
            "CPLUS_INCLUDE_PATH": "",
            "OBJC_INCLUDE_PATH": "",
            "GPR_RUNTIME_PATH": "",
        }

    def set_default_env(self) -> None:
        """Reset all env variables that can influence the build outcome."""
        for k, v in self.default_env.items():
            logger.debug('export %s="%s"', k, v)
            os.environ[k] = v

    @property
    def root_dir(self) -> str:
        """Root path of the sandbox.

        :raise SandBoxError: when the sandbox is not initialized
        """
        if self.__root_dir is None:
            raise SandBoxError(
                origin="root_dir", message="sandbox not loaded. Please call load()"
            )
        return self.__root_dir

    @root_dir.setter
    def root_dir(self, d: str) -> None:
        new_dir = os.path.realpath(d)
        if new_dir == self.__root_dir:
            return  # nothing to do

        self.__root_dir = new_dir

        # For each directory create an attribute containing its path
        for d in self.dirs:
            setattr(
                self, ("%s_dir" % d).replace(os.path.sep, "_"), os.path.join(new_dir, d)
            )

        # specs_dir path can be changed by a configuration in user.yaml
        user_yaml = os.path.join(new_dir, "user.yaml")
        if os.path.exists(user_yaml):
            with open(user_yaml) as f:
                self.user_config = yaml.safe_load(f)

            # Accept both specs_dir and module_dir key (in that order) to
            # get the path to the anod specification files
            specs_dir = (
                self.user_config.get("specs_dir", self.user_config.get("module_dir"))
                if self.user_config is not None
                else None
            )
            if specs_dir is not None:
                self.specs_dir = specs_dir
        else:
            self.__specs_dir = os.path.join(new_dir, "specs")

    @property
    def specs_dir(self) -> Optional[str]:
        """Return where to find anod specification files."""
        return self.__specs_dir

    @specs_dir.setter
    def specs_dir(self, d: str) -> None:
        """Set an alternate specs dir.

        :param d: directory where to find anod specification files
        """
        self.__specs_dir = d
        self.is_alternate_specs_dir = True
        logger.debug("using alternate specs dir %s", d)

    def create_dirs(self) -> None:
        """Create all required sandbox directories."""
        for d in self.dirs:
            mkdir(getattr(self, ("%s_dir" % d).replace(os.path.sep, "_")))

    def get_build_space(self, name: str, platform: Optional[str] = None) -> BuildSpace:
        """Get build space.

        :param name: build space name
        :param platform: platform name (if None use the default platform)

        :return: A BuildSpace object
        """
        if platform is None:
            platform = Env().platform
        return BuildSpace(root_dir=os.path.join(self.root_dir, platform, name))

    def dump_configuration(self) -> None:
        # Compute command line for call to e3-sandbox create. Ensure that the
        # paths are made absolute (path to sandbox, script).
        assert self.meta_dir is not None
        cmd_line = [sys.executable, os.path.abspath(__file__)]
        cmd_line += sys.argv[1:]
        sandbox_conf = os.path.join(self.meta_dir, "sandbox.yaml")
        with open(sandbox_conf, "w") as f:
            yaml.safe_dump({"cmd_line": cmd_line}, f)

    def get_configuration(self) -> dict:
        assert self.meta_dir is not None
        sandbox_conf = os.path.join(self.meta_dir, "sandbox.yaml")
        with open(sandbox_conf) as f:
            return yaml.safe_load(f)

    def write_scripts(self) -> None:
        from setuptools.command.easy_install import get_script_args

        assert self.bin_dir is not None

        # Retrieve sandbox_scripts entry points
        e3_distrib = get_distribution("e3-core")

        class SandboxDist:
            def get_entry_map(self, group):
                if group != "console_scripts":
                    return {}
                return e3_distrib.get_entry_map("sandbox_scripts")

            def as_requirement(self):
                return e3_distrib.as_requirement()

        for script in get_script_args(dist=SandboxDist()):
            script_name = script[0]
            script_content = script[1]
            target = os.path.join(self.bin_dir, script_name)
            rm(target)
            if not script_name.endswith(".exe"):
                script_content = script_content.replace(
                    "console_scripts", "sandbox_scripts"
                )
            with open(target, "wb") as f:
                if isinstance(script_content, str):
                    f.write(script_content.encode("utf-8"))
                else:
                    f.write(script_content)
            chmod("a+x", target)



# type: ignore
import logging
import os

import e3.log


def init_testsuite_env():
    """Initialize testsuite environment."""
    # Activate full debug logs
    e3.log.activate(level=logging.DEBUG, e3_debug=True)

    # Force UTC timezone
    os.environ["TZ"] = "UTC"
    os.environ["E3_ENABLE_FEATURE"] = "smtp_ssl"
    os.environ["E3_CONFIG"] = "/dev/null"
    # Ignore E3_HOSTNAME variable
    if "E3_HOSTNAME" in os.environ:
        del os.environ["E3_HOSTNAME"]


init_testsuite_env()


def pytest_addoption(parser):
    parser.addoption(
        "--ci", action="store_true", help="Tests are running on a CI server"
    )
    parser.addoption(
        "--requirement-coverage-report", help="Report requirement coverage"
    )



from __future__ import annotations

from typing import Callable, List, Optional


def process_exit_code(handle: int) -> Optional[int]:
    """Retrieve process exit code.

    Note that the method used ensure the process stays in a waitable state

    :param handle: process handle
    :return: the exit code if process is finished or None otherwise
    :raise: WindowsError in case the handle is invalid
    """
    from e3.os.windows.native_api import ProcessInfo, NT
    from ctypes import pointer, sizeof

    process_info = ProcessInfo.Basic()
    query_infoprocess: Callable = NT.QueryInformationProcess  # type: ignore
    status = query_infoprocess(
        handle,
        ProcessInfo.Basic.class_id,
        pointer(process_info),
        sizeof(process_info),
        None,
    )
    exit_code = process_info.exit_status

    if status < 0:
        raise OSError("invalid handle")
    elif exit_code == ProcessInfo.STILL_ACTIVE:
        return None
    else:
        return exit_code


def wait_for_objects(
    object_list: List[int], timeout: int = 0, wait_for_all: bool = False
) -> Optional[int]:
    """Wait until list of object are in signaled state.

    :param object_list: a list of handles
    :param timeout: maximum waiting time in seconds. If 0 then maximum waiting
        time is set to infinity
    :param wait_for_all: if True wait for all object to be signaled. If False
        wait for one object to be signaled
    :return: index in the object list of the signaled object or None in case
        of timeout
    :raise: WindowsError
    """
    from e3.os.windows.native_api import NT, Wait
    from ctypes.wintypes import HANDLE

    if timeout == 0:
        timeout = Wait.INFINITE
    else:
        timeout = int(timeout * 1000)

    size = len(object_list)
    handle_array = HANDLE * size
    handles = handle_array(*object_list)

    wait_for_multiples_objects: Callable = NT.WaitForMultipleObjects  # type: ignore

    object_index = wait_for_multiples_objects(size, handles, wait_for_all, timeout)
    if object_index == Wait.TIMEOUT:
        return None
    elif object_index == Wait.FAILED:  # defensive code
        raise OSError("error while waiting for objects")
    elif Wait.ABANDONED <= object_index < Wait.ABANDONED + size:  # defensive code
        return object_index - Wait.ABANDONED
    elif Wait.OBJECT <= object_index < Wait.OBJECT + size:
        return object_index
    else:  # defensive code
        raise OSError("unknown error while waiting for objects")



"""Main program initialization.

This package provides a class called Main used to initialize a python script
invoked from command line. The main goal is to ensure consistency in term of
interface, documentation and logging activities for all scripts using e3.

The script will support by default the following switches::

    -v|--verbose to enable verbose mode (a console logger is added)
    -h|--help    display command line help
    --log-file FILE
                 to redirect logs to a given file (this is independent from
                 verbose option
    --console-logs
                 disable color, progress bars, and redirect as much as
                 possible to stdout, starting lines with the given prefix

In addition, if the add_targets_options parameter is set to True
when instantiating an object of class Main, the following switches
will also be provided::

    --build      to set the build
    --host       to set the host
    --target     to set the target
"""


from __future__ import annotations

from argparse import ArgumentParser
import logging
import os
import signal
import sys

from typing import TYPE_CHECKING

import e3
import e3.log
from e3.env import Env

if TYPE_CHECKING:
    from typing import Optional, List
    from argparse import Namespace


class Main:
    """Class that implement argument parsing.

    :ivar args: list of positional parameters after processing options
    """

    def __init__(
        self,
        name: Optional[str] = None,
        platform_args: bool = False,
        argument_parser: Optional[ArgumentParser] = None,
    ):
        """Initialize Main object.

        :param name: name of the program (if not specified the filename without
            extension is taken)
        :param platform_args: add --build, --host, --target support
        :param argument_parser: the ArgumentParser to use for parsing
            command-line arguments (if not specified, an ArgumentParser will be
            created by Main)
        """
        # On UNIX set a signal handler for SIGTERM that will raise SystemExit
        # This is to let an e3 application enough time to perform
        # cleanup steps when killed by rlimit. rlimit first send a SIGTERM
        # then a SIGKILL 5 seconds later

        main = sys.modules["__main__"]

        if name is not None:
            self.name = name
        else:
            self.name = os.path.splitext(os.path.basename(main.__file__))[0]

        if argument_parser is None:
            argument_parser = ArgumentParser()

        log_group = argument_parser.add_argument_group(title="Logging arguments")
        log_group.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="make the log outputted on the console more verbose (this "
            "sets the log level to DEBUG)",
        )
        log_group.add_argument(
            "--log-file",
            metavar="FILE",
            default=None,
            help="store all the logs into the specified file",
        )
        log_group.add_argument(
            "--loglevel",
            default=logging.INFO,
            help="set the console log level",
            choices={
                "DEBUG": logging.DEBUG,
                "INFO": logging.INFO,
                "ERROR": logging.ERROR,
                "CRITICAL": logging.CRITICAL,
            },
        )
        log_group.add_argument(
            "--nocolor",
            default=False,
            action="store_true",
            help="disable color and progress bars",
        )
        log_group.add_argument(
            "--json-logs",
            default="json-logs" in os.environ.get("E3_ENABLE_FEATURE", "").split(","),
            action="store_true",
            help="enable JSON formatted logs. They can be activated as well by"
            " setting the env var E3_ENABLE_FEATURE=json-logs.",
        )
        log_group.add_argument(
            "--console-logs",
            metavar="LINE_PREFIX",
            help="disable color, progress bars, and redirect as much as"
            " possible to stdout, starting lines with the given prefix.",
        )

        if platform_args:
            plat_group = argument_parser.add_argument_group(title="Platform arguments")
            plat_group.add_argument(
                "--build",
                default=None,  # to force autodetection
                help="Set the build platform and build os version",
            )
            plat_group.add_argument(
                "--host",
                default=None,  # to force autodetection
                help="Set the host platform, host os version",
                metavar="HOST[,HOST_VERSION]",
            )
            plat_group.add_argument(
                "--target",
                default=None,  # to force autodetection
                help="Set the target platform, target os version, "
                "target machine, and target mode",
                metavar="TARGET[,TARGET_VERSION[,TARGET_MACHINE[," "TARGET_MODE]]]",
            )
            # We add a default to a fake option as a way to encode
            # the fact that this parser supports the standard
            # --build/host/target options. That way, after the parser
            # is used to evaluate the command-line arguments, we can
            # determine from the result whether the parser was supporting
            # the standard --host/target options or not, and then process
            # them if we did.
            #
            # To avoid clashes with user-defined options, we use a dest
            # name that is improbable in practice.
            argument_parser.set_defaults(e3_main_platform_args_supported=True)

        self.args: Optional[Namespace] = None
        self.argument_parser = argument_parser
        self.__log_handlers_set = False

        def sigterm_handler(sig, frame):  # unix-only
            """Automatically convert SIGTERM to SystemExit exception.

            This is done to give enough time to an application killed by
            rlimit to perform the needed cleanup steps
            :param sig: signal action
            :param frame: the interrupted stack frame
            """
            del sig, frame
            logging.critical("SIGTERM received")
            raise SystemExit("SIGTERM received")

        if sys.platform != "win32":  # unix-only
            signal.signal(signal.SIGTERM, sigterm_handler)

    def parse_args(
        self, args: Optional[List[str]] = None, known_args_only: bool = False
    ) -> None:
        """Parse options and set console logger.

        :param args: the list of positional parameters. If None then
            ``sys.argv[1:]`` is used
        :param known_args_only: does not produce an error when extra
            arguments are present
        """
        if known_args_only:
            self.args, _ = self.argument_parser.parse_known_args(args)
        else:
            self.args = self.argument_parser.parse_args(args)

        if self.args.nocolor:
            e3.log.pretty_cli = False

        if not self.__log_handlers_set:
            # First set level of verbosity
            if self.args.verbose:
                level = logging.DEBUG
            else:
                level = self.args.loglevel

            if self.args.console_logs:
                e3.log.console_logs = self.args.console_logs

            e3.log.activate(
                level=level,
                filename=self.args.log_file,
                json_format=self.args.json_logs,
                e3_debug=self.args.verbose > 1,
            )
            self.__log_handlers_set = True

        # Export options to env
        e = Env()
        e.main_options = self.args

        if hasattr(self.args, "e3_main_platform_args_supported"):
            e3.log.debug("parsing --build/--host/--target")
            # Handle --build, --host, and --target arguments
            e.set_env(self.args.build, self.args.host, self.args.target)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import e3.anod.error
import e3.diff
import e3.log
from e3.archive import create_archive
from e3.fs import VCS_IGNORE_LIST, mkdir, rm, sync_tree


if TYPE_CHECKING:
    from typing import Callable, Dict, Final, List, Literal, NoReturn, Optional, Union
    from e3.anod.spec import Anod

    PrepareSrcCB = Callable[[Dict[str, Dict[str, str]], str], None]
    ApplyPatchCB = Callable[[str, str, str], None]


class Package:
    """Describe and generate a binary package."""

    def __init__(
        self,
        prefix: str,
        publish: bool = False,
        version: Optional[Callable[[], str]] = None,
    ):
        """Create a binary package.

        :param prefix: prefix of the package to create, the name will be
            {prefix}-{version}-{platform}-bin.{exe,tar.gz}
            The version is by default set to 'unknown' and
            can be overridden by the version callback.
        :param publish: if True, publish the package (i.e. the package
            can be distributed to a customer).
        :param version: a callback returning the package version, if None the
            version is set to Anod.sandbox.build_version
        """
        self.prefix = prefix
        self.name = prefix + "-{version}-{platform}-bin"
        self.publish = publish
        self.version = version

    def pkg_name(self, anod_instance: Anod) -> str:
        """Return the final package filename.

        :param anod_instance: the Anod instance that creates the package
        :return: the name without extension of the package filename
        """
        if self.version is not None:
            version = self.version()
        else:
            version = "unknown"

        return self.name.format(version=version, platform=anod_instance.env.platform)

    def pkg_path(self, anod_instance: Anod) -> str:
        """Return the full path in which a package will be generated.

        :param anod_instance: the Anod instance that creates the package
        :return: the full path to the generated archive
        """
        return os.path.join(
            anod_instance.build_space.binary_dir, self.pkg_name(anod_instance) + ".zip"
        )

    def create_package(self, anod_instance: Anod) -> str:
        """Generate a package as a ZIP archive.

        :param anod_instance: the Anod instance that creates the package
        :return: the full path to the generated archive
        """
        pkg_name = self.pkg_name(anod_instance)
        pkg_path = self.pkg_path(anod_instance)

        # Reset binary dir
        rm(anod_instance.build_space.binary_dir, True)
        mkdir(anod_instance.build_space.binary_dir)

        # Create the zip archive
        create_archive(
            filename=os.path.basename(pkg_path),
            from_dir=anod_instance.build_space.pkg_dir,
            dest=os.path.dirname(pkg_path),
            from_dir_rename=pkg_name,
        )
        return pkg_path


class Source:
    """Source package."""

    def __init__(
        self,
        name: str,
        publish: bool,
        dest: Optional[str] = None,
        unpack_cmd: Optional[Callable] = None,
        remove_root_dir: bool = True,
        ignore: None = None,
    ):
        """Create a new source object.

        Register a source package and describe where it should be installed.

        :param name: source package short alias (the full package name will
            be written in the sources metadata)
        :param publish: whether the source package should be distributed with
            the official releases
        :param dest: subdirectory of the local `src_dir` where to extract the
            source package
        :param unpack_cmd: the function to call to install the source package
            See gnatpython.fileutils.unpack_archive for more information
        :param remove_root_dir: if True (default) the root directory of the
            archive is ignored
        :param ignore: unused during instantiation, kept for backward
            compatibiliy.
        """
        del ignore
        self.name = name
        self.publish = publish
        if dest is None:
            dest = ""
        else:
            if os.path.isabs(dest) or os.pardir in dest:
                raise e3.anod.error.SpecError("dest should be relative path")
        self.dest = dest
        self.unpack_cmd = unpack_cmd
        self.remove_root_dir = remove_root_dir
        self.builder: Optional[SourceBuilder] = None
        self.other_sources: List[Source] = []

    def set_builder(self, builder_function: SourceBuilder) -> None:
        """Set the SourceBuilder associated to this Source object."""
        self.builder = builder_function

    def set_other_sources(self, other_sources: List[Source]) -> None:
        """Get the list of other sources to compute ``ignore`` property."""
        self.other_sources = other_sources

    @property
    def ignore(self) -> List[str]:
        """Return list of paths to ignore when installing the source.

        By default, a source package is first unpacked and then install
        in the `dest` directory by using e3.fs.sync_tree. The ignore
        property returns a value that can be passed to e3.fs.sync_tree.
        """
        ignore_list = []
        for other_source in self.other_sources:
            if other_source.name != self.name and other_source.dest:
                ignore_path = os.path.relpath(other_source.dest, self.dest)
                if not ignore_path.startswith(os.pardir):
                    ignore_list.append(f"/{ignore_path}")
        return ignore_list


class SharedSource(Source):
    """Shared sources are installed directly in the sandbox.

    This is useful when you need to use the same sources for different
    modules/platforms and you don't need to write in the sources
    directory
    """

    pass


class SourceBuilder:

    DEFAULT_PATCH_CMD: Final = 1

    def __init__(
        self,
        name: str,
        fullname: Callable[[], str],
        checkout: Optional[List[str]],
        prepare_src: Optional[PrepareSrcCB] = None,
        apply_patch: Optional[Union[Literal[1], ApplyPatchCB]] = None,
        kind: str = "source",
    ):
        """Define a builder for the source package.

        :param name: short name of the package (as used by build_source_list
            and test_source_list).
        :param fullname: callback providing the full package name
        :param checkout: list of repository names needed to build the
            source package
        :param prepare_src: a function (repo_dict, dest) where
            repo_dict is a dictionary of {repo_name: working directory}
            dest is the path to the dest directory
            If None, just rsync from working directory to dest
            Note that it is supported only if the SourceBuilder object has
            only one Repository object. The repository working directories
            should be considered read only
        :param apply_patch: a function(repository, patch_file, dest) where
            repository is the name of the repository used to create a patch
            patch_file is a patch to apply
            dest is the path to the dest directory
            If apply_patch is set to DEFAULT_PATCH_CMD, use the default
            apply_patch command
        :param kind: type of source 'source' or 'thirdparty'
        """
        self.name = name
        self.checkout = checkout if checkout is not None else []
        self.repositories = {}  # type: ignore
        self.kind = kind
        self.from_spec = None
        self.source_query: Optional[Dict[str, Optional[str]]] = None
        self.__fullname = fullname
        self.__prepare_src = prepare_src
        self.__apply_patch = apply_patch

    def fullname(self, *args, **kwargs):
        # ??? add support for the GPL mode
        return self.__fullname(*args, **kwargs)

    @property
    def prepare_src(self) -> Optional[PrepareSrcCB]:
        """Return the callback used to prepare the source package.

        :return: the callback or None if no prepare_src is defined
        :raise: SpecError
        """
        if self.__prepare_src is not None:
            return self.__prepare_src

        # Else provide a default function if we have exactly 1 checkout
        if not self.checkout:
            raise e3.anod.error.SpecError(
                "no checkout associated to the builder %s" % self.name, "prepare_src"
            )
        elif len(self.checkout) > 1:
            raise e3.anod.error.SpecError(
                "more than 1 checkout and no prepare_src function"
                " given for %s" % self.name
            )

        # Set default function (a basic sync_tree call) that ignore
        # .svn, .git, .cvs, .cvsignore and .gitignore files
        def default_prepare_src(repos: Dict[str, Dict[str, str]], dest: str) -> None:
            sync_tree(
                list(repos.values())[0]["working_dir"], dest, ignore=VCS_IGNORE_LIST
            )

        return default_prepare_src

    @property
    def apply_patch(self) -> ApplyPatchCB:
        """Return the callback used to apply a patch.

        :return: the callback
        """

        def default_apply_patch(_: str, patch_file: str, dest: str) -> None:
            """Apply a patch file using e3.diff.patch."""
            e3.log.debug("applying patch %s on %s", patch_file, dest)
            e3.diff.patch(patch_file, dest)

        def no_apply_patch(r: str, p: str, d: str) -> NoReturn:
            """Raise an error (no apply_patch function defined)."""
            # Unused parameters
            del r, p, d
            raise e3.anod.error.AnodError(
                "no apply_patch function defined in SourceBuilder %s" % self.name,
                "apply_path",
            )

        if self.__apply_patch == self.DEFAULT_PATCH_CMD or (
            self.prepare_src is not None
            and self.__apply_patch is None
            and len(self.checkout) == 1
        ):
            # Either the default patch function is forced or
            # a package can be created with Anod, no patch function is
            # provided and there is only a single repository to patch
            return default_apply_patch
        else:
            if self.__apply_patch is None:
                return no_apply_patch
            else:
                return self.__apply_patch


class UnmanagedSourceBuilder(SourceBuilder):
    """Source builder for sources not managed by anod."""

    @property
    def prepare_src(self) -> None:
        """Do not create source package."""
        return None


class ThirdPartySourceBuilder(UnmanagedSourceBuilder):
    """SourceBuilder for thirdparty packages."""

    def __init__(self, name: str):
        """Create a SourceBuilder for a third party package.

        :param name: full package name (including extension)
        """
        super().__init__(
            name=name,
            fullname=lambda: name,
            checkout=None,
            apply_patch=self.DEFAULT_PATCH_CMD,
            kind="thirdparty",
        )


class ExternalSourceBuilder(UnmanagedSourceBuilder):
    """SourceBuilder to reference sources produced outside the setup."""

    def __init__(
        self,
        name: str,
        bid: Optional[str] = None,
        setup: Optional[str] = None,
        date: Optional[str] = None,
        query_name: Optional[str] = None,
    ):
        """Initialize ExternalSourceBuilder.

        :param name: source name
        :param bid: build id. If not None then setup and date parameters
            are ignored
        :param setup: setup in which the source has been created. If bid
            is None then this parameter is mandatory
        :param date: date of the required source
        :param query_name: name of the source in 'setup'. Defaults to 'name'
            if None.
        """
        super().__init__(name=name, fullname=lambda: name, checkout=None, kind="source")
        if query_name is None:
            query_name = name
        self.source_query = {
            "name": query_name,
            "bid": bid,
            "setup": setup,
            "date": date,
        }



from __future__ import annotations

from typing import TYPE_CHECKING

import abc
import time

if TYPE_CHECKING:
    from typing import Any

DEFAULT_TIMEOUT = 3600 * 24


class Cache(metaclass=abc.ABCMeta):
    def __init__(self, cache_configuration: Any):
        self.cache_configuration = cache_configuration

    @abc.abstractmethod
    def get(self, uid: str, default: Any = None) -> Any:
        """Fetch a given resource from the cache.

        If the resource does not exist, return default, which itself defaults
        to None.
        :param uid: the resource uid
        :param default: the default value if not found
        """
        pass  # all: no cover

    def get_expiry_time(self, timeout: int = DEFAULT_TIMEOUT) -> float:
        """Return the expiry time based upon the provided timeout.

        :param timeout: timeout
        """
        return time.time() + timeout

    def has_resource(self, uid: str) -> bool:
        """Return True if the uid is in the cache and has not expired.

        :param uid: the resource uid
        """
        return self.get(uid) is not None

    @abc.abstractmethod
    def set(self, uid: str, value: Any, timeout: int = DEFAULT_TIMEOUT) -> bool:
        """Set a value in the cache.

        :param uid: the cache entry uid
        :param value: the object to cache
        :param timeout: timeout to use for caching this value, otherwise the
            default cache timeout will be used.
        :return: True if the value is set, False in case of failure
        """
        pass  # all: no cover

    @abc.abstractmethod
    def delete(self, uid: str) -> None:
        """Delete a resource from the cache.

        Do nothing if the uid does not exist.
        :param uid: the resource uid
        """
        pass  # all: no cover

    @abc.abstractmethod
    def clear(self) -> None:
        """Remove *all* values from the cache at once."""
        pass  # all: no cover

    def close(self) -> None:
        """Close the cache connection."""
        pass  # all: no cover

    def __contains__(self, uid: str) -> bool:
        """Return True if the resource is in the cache and has not expired.

        :param uid: the resource uid
        """
        return self.has_resource(uid)



import sys

import e3.os.process

import pytest

if sys.platform == "win32":
    import e3.os.windows.native_api


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_unicode_string_preallocation():
    s = e3.os.windows.native_api.UnicodeString(max_length=10)
    assert len(s) == 0, "preallocated unicode string has still a 0 length"



import os
import sys

import e3.env
import e3.fs
import e3.os.process
import e3.platform

import pytest


def test_autodetect():
    sys_platform = (
        sys.platform.replace("linux2", "linux")
        .replace("win32", "windows")
        .replace("aix7", "aix")
        .replace("sunos5", "solaris")
    )
    assert sys_platform in str(e3.platform.Platform.get())

    assert sys_platform in e3.env.Env().build.platform

    b = e3.env.BaseEnv()
    b.set_build("x86-linux", "rhES7")
    assert b.build.platform == "x86-linux"
    assert sys_platform in e3.env.Env().build.platform

    assert "--build=x86-linux,rhES7" in b.cmd_triplet()

    b.set_host("x86_64-linux", "rhES7")
    assert "--build=x86-linux,rhES7" in b.cmd_triplet()
    assert "--host=x86_64-linux,rhES7" in b.cmd_triplet()
    assert b.get_attr("build.os.version") == "rhES7"


def test_platform():
    e = e3.env.BaseEnv()
    e.set_host("x86-linux")
    assert e.platform == "x86-linux"
    e.set_target("x86-solaris")
    assert e.platform == "x86-solaris-linux"


def test_is_canadian():
    e = e3.env.BaseEnv()
    e.set_build("sparc-solaris")
    assert not e.is_canadian
    e.set_host("sparc64-solaris")
    assert not e.is_canadian
    e.set_host("x86-windows")
    assert e.is_canadian


def test_set_host():
    e = e3.env.BaseEnv()
    e.set_build("x86-linux")

    # check that set_host reset target to host
    e.set_target("x86-windows")
    e.set_host("x86_64-linux")
    assert e.target == e.host

    # check special value 'build'
    e.set_host("build")
    assert e.build == e.host
    assert e.target == e.build

    # check no value passed (should be the same as host='build')
    e.set_host("x86_64-linux")
    e.set_host()
    assert e.host == e.build
    assert e.target == e.build

    # check special value 'target'
    e.set_target("x86-windows")
    e.set_host("target")
    assert e.host == e.target


def test_set_target():
    e = e3.env.BaseEnv()
    e.set_build("x86-linux")
    e.set_host("x86_64-linux")

    # check 'build' special value
    e.set_target("build")
    assert e.is_cross
    assert e.is_canadian

    e.set_target("host")
    assert e.host == e.target

    e.set_target()
    assert e.host == e.target


def test_set_env():
    e = e3.env.BaseEnv()
    e.set_env("x86-linux,rhEs5", "x86_64-linux,debian7", "x86-windows,2008")
    assert e.build.platform == "x86-linux"
    assert e.host.platform == "x86_64-linux"
    assert e.target.platform == "x86-windows"

    # keep the current target
    e.set_env("x86-linux,rhEs5", "x86_64-linux,debian7", "target")
    assert e.target.platform == "x86-windows"

    # None means host, replace the target by the previous host value
    e.set_env("x86-linux,rhEs5", "x86_64-linux,debian7", None)
    assert e.target.platform == "x86_64-linux"

    # replace the target by the previous build value
    e.set_env("x86-linux,rhEs5", "x86_64-linux,debian7", "build")
    assert e.target.platform == "x86-linux"

    # replace the target by the previous build host
    e.set_env("x86-linux,rhEs5", "x86_64-linux,debian7", "host")
    assert e.target.platform == "x86_64-linux"


def test_cmd_triplet():
    if e3.env.Env().build.platform == "x86-linux":
        build_platform = "x86_64-linux,rhES5"
    else:
        build_platform = "x86-linux,rhES5"
    e = e3.env.BaseEnv()
    e.set_env(build_platform, "x86_64-linux,debian7", "x86-windows,2008")
    cmd_options = e.cmd_triplet()
    assert len(cmd_options) == 3
    assert cmd_options[0] == "--build=%s" % build_platform
    assert cmd_options[1] == "--host=x86_64-linux,debian7"
    assert cmd_options[2].startswith("--target=x86-windows,2008")


def test_get_attr():
    e = e3.env.BaseEnv()
    e.set_env("x86-linux,rhES5", "x86_64-linux,debian7", "x86-windows,2008")
    assert e.get_attr("host.os.name") == "linux"
    assert e.get_attr("host.os.name2", default_value="hello") == "hello"
    assert e.get_attr("host.cpu.bits", forced_value="gotit") == "gotit"

    e.my_attr = 3
    assert e.get_attr("my_attr") == 3

    e.my_attr = None
    assert e.get_attr("my_attr", default_value=4) == 4

    assert e.my_attr is None

    with pytest.raises(AttributeError):
        print(e.does_not_exist)


def test_add_path():
    e = e3.env.Env()
    saved_path = os.environ["PATH"]
    e.store()
    e.add_path("/dummy_for_test")
    assert os.environ["PATH"].startswith("/dummy_for_test")
    e.add_path("/dummy_for_test", append=True)
    assert os.environ["PATH"].endswith("/dummy_for_test")
    e.restore()
    assert saved_path == os.environ["PATH"]


def test_add_dll_path():
    e = e3.env.Env()
    saved_path = os.environ.get(e.dll_path_var)
    e.store()
    e.add_dll_path("/dummy_for_test")
    assert os.environ[e.dll_path_var].startswith("/dummy_for_test")
    e.add_dll_path("/dummy_for_test", append=True)
    assert os.environ[e.dll_path_var].endswith("/dummy_for_test")
    e.restore()
    assert saved_path == os.environ.get(e.dll_path_var)


def test_discriminants():
    e = e3.env.Env()
    e.store()
    assert "native" in e.discriminants
    e.set_target("arm-elf")
    assert "native" not in e.discriminants

    e.set_build("x86-windows")
    assert "NT" in e.discriminants
    e.restore()


def test_tmp():
    e = e3.env.Env()
    current_dir = os.getcwd()
    os.environ["TMPDIR"] = current_dir
    assert e.tmp_dir == current_dir


def test_to_dict():
    assert e3.env.Env().to_dict()["is_cross"] is False


def test_store():
    c = e3.env.Env()

    c.abc = "foo"

    c.store()

    c.abc = "bar"

    c.restore()

    assert c.abc == "foo"

    c.store()
    c.abc = "one"
    c.store("store")
    c.abc = "two"
    c.restore("store")
    assert c.abc == "one"
    c.restore()
    assert c.abc == "foo"

    # calling .restore() when there is nothing to restore is a no-op
    # this test is run through env_protect (defined in conftest.py)
    # so Env().store() has already been called, call it twice here
    c.restore()
    c.restore()

    assert not hasattr(c, "abc")


def test_from_platform_name():
    e = e3.env.BaseEnv.from_platform_name("arm-linux-linux")
    assert e.target.platform == "arm-linux"
    assert e.build.platform == "x86-linux"
    e = e3.env.BaseEnv.from_platform_name("arm-linux-linux64")
    assert e.target.platform == "arm-linux"
    assert e.build.platform == "x86_64-linux"
    e = e3.env.BaseEnv.from_platform_name("x86_64-linux-darwin")
    assert e.target.platform == "x86_64-linux"
    assert e.build.platform == "x86_64-darwin"
    assert e.is_cross
    e = e3.env.BaseEnv.from_platform_name("x86_64-linux")
    assert e.target.platform == "x86_64-linux"
    assert e.build.platform == "x86_64-linux"
    assert not e.is_cross
    e = e3.env.BaseEnv.from_platform_name("avr-elf-solaris")
    assert e.target.platform == "avr-elf"
    assert e.build.platform == "sparc-solaris"
    assert e.is_cross

    e = e3.env.BaseEnv.from_platform_name("what-linux-linux")
    assert e is None


def test_copy():
    e = e3.env.BaseEnv()
    new_e = e.copy(target="arm-elf")

    assert new_e.target.platform == "arm-elf"


def test_build_os_propagation():
    """Ensure that OS version is preserved when build platform is changed."""
    winenv = e3.env.BaseEnv(
        build=e3.platform.Platform.get(
            platform_name="x86-windows", version="2012", machine="mywindows"
        )
    )

    winenv.set_env(build="x86_64-windows")

    assert winenv.build.os.version == "2012"



import os

from e3.anod.error import SandBoxError
from e3.anod.loader import AnodSpecRepository
from e3.fs import sync_tree
from e3.os.process import Run

import pytest


class TestLoader:

    spec_dir = os.path.join(os.path.dirname(__file__), "data")
    spec2_dir = os.path.join(os.path.dirname(__file__), "data2")

    def test_spec_does_not_exist(self):
        with pytest.raises(SandBoxError) as err:
            AnodSpecRepository("/foo/bar")

        assert str(err.value).startswith("spec directory /foo/bar does not exist")

    def test_spec_loader1(self):
        spec_repo = AnodSpecRepository(self.spec_dir)
        s = spec_repo.load("loader1")
        assert s.name == "loader1"

    def test_spec_loader2(self):
        spec_repo = AnodSpecRepository(self.spec_dir)

        with pytest.raises(SandBoxError) as err:
            spec_repo.load("loader2")
        assert str(err.value).startswith("load: cannot find Anod subclass in")

    def test_invalid_spec(self):
        """Ensure that loading an invalid spec result in a SandboxError."""
        spec_repo = AnodSpecRepository(self.spec_dir)
        with pytest.raises(SandBoxError) as err:
            spec_repo.load("invalid_spec")

        assert "invalid spec code" in str(err.value)

    def test_spec_loader_prolog(self):
        spec_repo = AnodSpecRepository(self.spec_dir, spec_config=True)
        anod_class = spec_repo.load("prolog_test")

        # We should be able to load a spec twice
        anod_class = spec_repo.load("prolog_test")

        anod_instance = anod_class("prolog_test", "", "build")
        assert anod_instance.prolog_test, "prolog not executed properly"

    def test_spec_inheritance(self):
        """Load a spec that inherit from another spec."""
        spec_repo = AnodSpecRepository(self.spec_dir)
        anod_class = spec_repo.load("child")
        anod_instance = anod_class("load", "", "build")
        assert anod_instance.parent_info == "from_parent"

    def test_multiple_spec_repository(self):
        """Ensure that spec function is context dependent."""
        spec_repo = AnodSpecRepository(self.spec_dir)
        spec2_repo = AnodSpecRepository(self.spec2_dir)
        anod_class = spec_repo.load("child")
        anod_instance = anod_class("load", "", "build")
        assert anod_instance.parent_info == "from_parent"
        anod_class2 = spec2_repo.load("child")
        anod_instance2 = anod_class2("load", "", "build")
        assert anod_instance2.parent_info == "from_parent2"

    def test_load_all(self):
        spec_repo = AnodSpecRepository(self.spec_dir)
        with pytest.raises(SandBoxError):
            spec_repo.load_all()

        spec_repo = AnodSpecRepository(self.spec2_dir)
        spec_repo.load_all()
        assert "parent" in spec_repo
        assert "child" in spec_repo
        assert "unknown" not in spec_repo

    def test_load_config(self):
        spec_repo = AnodSpecRepository(self.spec_dir)
        spec_repo.api_version = "1.4"
        anod_class = spec_repo.load("withconfig")
        anod_instance = anod_class("", "build")

        # See comments in tests/tests_e3/anod/data/withconfig.anod

        assert anod_instance.test1() == 9
        with pytest.raises(KeyError) as err:
            anod_instance.test2()
        assert "foo" in str(err.value)

        assert list(anod_instance.test3()) == ["case_foo"]

    def test_load_config_api_1_5(self):
        sync_tree(self.spec_dir, "new_spec_dir")
        Run(["e3-sandbox", "migrate", "1.5", "new_spec_dir"], output=None)
        spec_repo = AnodSpecRepository("new_spec_dir")
        spec_repo.api_version = "1.5"
        anod_class = spec_repo.load("withconfig")
        anod_instance = anod_class("", "build")
        assert anod_instance.test1() == 9
        assert anod_instance.test_suffix() == 42

    def test_sandbox_migrate_unknown_api(self):
        p = Run(["e3-sandbox", "migrate", "0.2", "foo"])
        assert "Only 1.5 is supported" in p.out

    def test_getitem_without_buildspace(self):
        """Without a build space PKG_DIR returns 'unknown'."""
        spec_repo = AnodSpecRepository(self.spec_dir)
        anod_class = spec_repo.load("parent")
        anod_instance = anod_class("", "build")
        assert anod_instance["PKG_DIR"] == "unknown"



import e3.electrolyt.host as host


def test_host_db():
    db = host.HostDB()
    db.add_host(
        hostname="computer1",
        platform="x86_64-linux",
        version="rhes5",
        data_center="dc832",
    )

    assert db.hostnames == ["computer1"]
    assert db.get("computer1").data_center == "dc832", db.get("computer1").__dict__
    assert db["computer1"].platform == "x86_64-linux"


def test_host_db_yaml():
    with open("db.yaml", "w") as f:
        f.write(
            "computer2:\n"
            "   build_platform: x86-windows\n"
            "   build_os_version: 2008R2\n"
            "   data_center: dc993\n\n"
            "computer3:\n"
            "   build_platform: x86_64-darwin\n"
            "   build_os_version: 16.3\n"
            "   data_center: dcmac\n"
        )

    db = host.HostDB(filename="db.yaml")
    assert set(db.hostnames) == {"computer2", "computer3"}
    assert db["computer3"].platform == "x86_64-darwin"


def test_host_db_yaml_alias():
    with open("db.yaml", "w") as f:
        f.write(
            "computer2: &computer2_alias\n"
            "   build_platform: x86-windows\n"
            "   build_os_version: 2008R2\n"
            "   data_center: dc993\n\n"
            "computer3: *computer2_alias\n"
        )

    db = host.HostDB(filename="db.yaml")
    assert set(db.hostnames) == {"computer2", "computer3"}
    assert db["computer3"].platform == "x86-windows"



import e3.anod.deps

import pytest


def test_buildvar():
    with pytest.raises(AssertionError):
        e3.anod.deps.BuildVar("foo", [1, 2])

    b = e3.anod.deps.BuildVar("foo", "bar")
    assert b.kind == "var"
    assert str(b) == "foo=bar"


def test_dependency():
    d = e3.anod.deps.Dependency("dep", product_verison="20.0855369232", build="default")

    parent_env = e3.env.BaseEnv(build=e3.platform.Platform.get("arm-ios"))

    class MockAnod:
        pass

    parent_anod_instance = MockAnod()
    parent_anod_instance.env = parent_env

    defaultenv = e3.env.BaseEnv(build=e3.platform.Platform.get("arm-linux"))

    assert (
        d.env(parent_anod_instance, default_env=defaultenv).build.platform
        == defaultenv.build.platform
    )
    assert (
        d.env(parent_anod_instance, default_env=defaultenv).host.platform
        == defaultenv.build.platform
    )
    assert (
        d.env(parent_anod_instance, default_env=defaultenv).target.platform
        == defaultenv.build.platform
    )


def test_dependency2():
    d = e3.anod.deps.Dependency("dep", product_verison="20.0855369232", host="default")

    parent_env = e3.env.BaseEnv(build=e3.platform.Platform.get("arm-ios"))

    class MockAnod:
        pass

    parent_anod_instance = MockAnod()
    parent_anod_instance.env = parent_env

    defaultenv = e3.env.BaseEnv(build=e3.platform.Platform.get("arm-linux"))

    assert (
        d.env(parent_anod_instance, default_env=defaultenv).host.platform
        == defaultenv.build.platform
    )
    assert (
        d.env(parent_anod_instance, default_env=defaultenv).target.platform
        == defaultenv.build.platform
    )


def test_dependency3():
    d = e3.anod.deps.Dependency(
        "dep", product_verison="20.0855369232", target="default"
    )

    parent_env = e3.env.BaseEnv(build=e3.platform.Platform.get("arm-ios"))

    class MockAnod:
        pass

    parent_anod_instance = MockAnod()
    parent_anod_instance.env = parent_env

    defaultenv = e3.env.BaseEnv(build=e3.platform.Platform.get("arm-linux"))

    assert (
        d.env(parent_anod_instance, default_env=defaultenv).target.platform
        == defaultenv.build.platform
    )



from __future__ import annotations

import abc
import json
import os
import sys
import time
import uuid
from typing import TYPE_CHECKING

import e3.env
import e3.hash
import e3.log
import stevedore
from e3.date import timestamp_as_string
from e3.error import E3Error
from e3.fs import mkdir

logger = e3.log.getLogger("event")

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Optional, Tuple, Union


def unique_id() -> str:
    """Return a random globally unique id.

    :return: an id
    """
    # Using clock_seq ensures that the uid of the event has at least
    # microsecond precision.
    return str(uuid.uuid1(clock_seq=int(1000 * time.time())))


class Event:
    """Event class for notifying external services.

    An event is composed of:

      * some data organized as key value dict
      * some attachments

    By default the data will contain the following keys:
      * name: the name of the event
      * uid: a global unique id
      * begin_time: time at which the event was created
      * end_time: time at which the event was closed
    """

    def __init__(self, name: str, uid: Optional[str] = None, **kwargs: Dict[str, str]):
        """Initialize an Event object.

        :param name: name of the event (e.g. e3_test)
        :param uid: unique identifier. If not given then an automatic uuid is
            computed
        :param kwargs: additional key value pairs
        """
        # Internal attributes. All other attributes are store in _data. By
        # using this construct we ensure the used cannot modify directly these
        # internal.
        object.__setattr__(self, "_attachments", {})
        object.__setattr__(self, "_data", {})
        object.__setattr__(self, "_closed", False)
        object.__setattr__(self, "_formatters", {})

        self.uid = uid if uid is not None else unique_id()
        self.name = name
        self.begin_time = time.time()
        self.end_time: Optional[float] = None

        for key, value in list(kwargs.items()):
            self._data[key] = value  # type: ignore

        self.set_formatter("begin_time", self.format_date)
        self.set_formatter("end_time", self.format_date)

    def __enter__(self) -> Event:
        return self

    def __exit__(self, _type, _value, _tb):
        self.close()

    def set_formatter(self, key: str, fun: Callable[[str, Any], dict]) -> None:
        """Add a formatter for a given key. (see as_dict).

        :param key: the event attribute to format
        :param fun: a function that takes the key and the associated value
            and return a dict
        """
        self._formatters[key] = fun  # type: ignore

    def __setattr__(self, name: str, value: str) -> None:
        """Store all attributes in the self._data dict."""
        # Once the event is closed disallow attributes modifications
        if self._closed:
            raise EventError(f"event {self.name} ({self.uid}) closed")
        self._data[name] = value  # type: ignore

    def __getattr__(self, name: str) -> str:
        """Attributes are retrieved in the _data internal dict."""
        try:
            return self._data[name]  # type: ignore
        except KeyError as e:
            raise AttributeError(e).with_traceback(sys.exc_info()[2])

    def get_attachments(self) -> Dict[str, Tuple[str, str]]:
        """Return the list of attachments.

        :return: a list of tuple (path, sha1(path))
        """
        return self._attachments  # type: ignore

    def attach_file(self, path: str, name: str = "log") -> None:
        """Attach log file to the event.

        When the event will be submitted, the log file will be attached.
        Note that some notification backend might cut or reject the attachment
        if too big.
        :param path: path to a log file
        :param name: name of the file to attach, by default 'log'
        """
        if self._closed:
            raise EventError(f"event {self.name} ({self.uid}) closed")
        self._attachments[name] = (path, e3.hash.sha1(path))  # type: ignore

    def close(self) -> None:
        """Close the event. Once done it is not supposed to be modified.

        Calling the method close() allow using it with
        contexlib.closing()
        """
        if not self._closed:
            self.end_time = time.time()
            object.__setattr__(self, "_closed", True)

    def format_date(self, key: str, value: Union[float, str]) -> Dict[str, str]:
        """Format timestamp fields.

        :param key: the data key
        :param value: a timestamp
        :return: a dict associating the original key to a human readable date
        """
        if isinstance(value, float):
            return {key: timestamp_as_string(value)}
        else:
            return {key: value}

    def as_dict(self) -> Dict[str, str]:
        """Convert the event data into a dict that can be serialized as json.

        For each key, value of the _data dict by default add them into the
        returned dict (default) or use a formatter that return a dict used to
        update the result.

        :return: a dict
        """
        result = {}
        for key, value in list(self._data.items()):  # type: ignore
            if key in self._formatters:
                d = self._formatters[key](key, value)  # type: ignore
                result.update(d)
            else:
                result[key] = value
        return result

    def dump(self, event_dir: str) -> str:
        """Dump the event into a json file.

        :param event_dir: directory in which the json is dumped
        :return: json file location
        """
        result = {
            "data": self.as_dict(),
            "attachments": self._attachments,
            "closed": self._closed,
        }
        mkdir(event_dir)
        json_filename = os.path.join(event_dir, f"{self.uid}-{unique_id()}.json")
        with open(json_filename, "w") as fd:
            json.dump(result, fd)
        return json_filename

    @classmethod
    def load(cls, json_filename: str) -> Event:
        """Retrieve an event from a JSON file.

        :param json_filename: file from which event is loaded
        :return: an event
        """
        with open(json_filename) as fd:
            event_dict = json.load(fd)

        result = Event(name="unknown")
        object.__setattr__(result, "_attachments", event_dict["attachments"])
        object.__setattr__(result, "_data", event_dict["data"])
        object.__setattr__(result, "_closed", event_dict["closed"])
        return result


class EventHandler(metaclass=abc.ABCMeta):
    """Interface to implement in order to be able to send events.

    One method needs to be implemented by a new EventManager:

    - send_event: method sending an event to an external service
    """

    @abc.abstractmethod
    def send_event(self, event: Event) -> bool:
        """Send an event.

        :param event: an Event
        :return: True on success, False otherwise
        """
        pass  # all: no cover

    @classmethod
    def decode_config(cls, config_str: str) -> dict:
        """Decode a config string into a dict.

        :param config_str: the string containing configuration information
        :return: a dict that can be used as **kwargs for the handler
            initialization method
        """
        return {}

    def encode_config(self) -> str:
        """Encode the handler configuration into a string.

        This default implementation can be used for handlers that do not
        need any configuration parameters (i.e: for which __init__ does not
        take any parameter apart for self).

        :return: a string that contains the current handler configuration.
            The string should not contain the '|' character.
        """
        return ""


class EventError(E3Error):
    pass


class EventManager:
    """Manage a set of handlers that will be used to send events."""

    def __init__(self) -> None:
        """Initialize a manager."""
        self.handlers: Dict[str, EventHandler] = {}

    def send_event(self, event: Event) -> bool:
        """Send an event to using all registered handlers.

        :param event: an event
        :return: True if the event was sent successfully to all handlers
        """
        status = True

        for handler in list(self.handlers.values()):
            handler_status = handler.send_event(event)
            if not handler_status:
                status = False

        return status

    def send_event_from_file(self, filename: str) -> bool:
        """Send an event from a dumped event.

        :param filename: path to the json file containing the event
        :return: True if the event was sent successfully to all handlers
        """
        e = Event.load(filename)
        return self.send_event(e)

    def get_handler(self, name: str) -> Callable[..., EventHandler]:
        """Get an handler class by name.

        Available handler classes are registered using the e3.event.handler
        entry_points in your setup.py

        :param name: handler name
        :return: an handler class
        """
        return stevedore.DriverManager("e3.event.handler", name).driver

    def add_handler(self, name: str, *args: Any, **kwargs: Any) -> None:
        """Add an handler instance to the manager.

        args and kwargs are passed to the handler __init__ method

        :param name: the handler name
        """
        logger.info("Add handler %s (%s %s)", name, args, kwargs)
        self.handlers[name] = self.get_handler(name)(*args, **kwargs)

    def load_handlers_from_env(self, var_name: str = "E3_EVENT_HANDLERS") -> None:
        """Add handlers by decoding an env variable.

        The variable value should have the following format:

            handler_name1=some_value|handler_name2=some_value|handler_name3

        The value associated with each handler is passed to the handler method
        decode_config. In most cases the user do not need to create this value
        manually but called handler_config_as_env method to create the
        variable. The main goal of this function is to share EventManager
        configuration among several processes

        :param var_name: the name of the variable
        """
        handler_cfg_str = os.environ.get(var_name, "")
        handler_cfg_dict: Dict[str, str] = dict(
            [
                el.split("=", 1)  # type: ignore
                if "=" in el
                else (el, "")
                for el in handler_cfg_str.split("|")
            ]
        )
        for handler_name, handler_config in list(handler_cfg_dict.items()):
            handler = self.get_handler(handler_name)
            logger.info("Add handler %s (%s)", handler_name, handler_config)
            self.handlers[handler_name] = handler(
                **handler.decode_config(handler_config)  # type: ignore
            )

    def handler_config_as_env(self, var_name: str = "E3_EVENT_HANDLERS") -> None:
        """Add handlers by decoding an env variable.

        :param var_name: the name of the variable containing the handler
            configurations
        """
        result = []
        for handler_name, handler in list(self.handlers.items()):
            config_str = handler.encode_config()
            assert "|" not in config_str
            result.append(f"{handler_name}={config_str}")
        os.environ[var_name] = "|".join(result)


# Declare the default manager
default_manager = EventManager()


def send_event(event: Event) -> bool:
    """Send event using default manager.

    See EventManager.send_event

    :param event: an event
    """
    return default_manager.send_event(event)


def send_event_from_file(filename: str) -> bool:
    """Send event from a file using default manager.

    See EventManager.send_event_from_file
    """
    return default_manager.send_event_from_file(filename)


def add_handler(name: str, *args: Any, **kwargs: Any) -> None:
    """Add handler in the default manager.

    See EventManager.add_handler
    """
    return default_manager.add_handler(name, *args, **kwargs)


def load_handlers_from_env(var_name: str = "E3_EVENT_HANDLERS") -> None:
    """Load handlers to default manager using env var.

    See EventManager.load_handlers_from_env

    :param var_name: the name of the variable containing the configuration
    """
    return default_manager.load_handlers_from_env(var_name=var_name)


def handler_config_as_env(var_name: str = "E3_EVENT_HANDLERS") -> None:
    """Export default manager handler configurations into env.

    See EventManager.handler_config_as_env

    :param var_name: the name of the variable
    """
    return default_manager.handler_config_as_env(var_name=var_name)



import ast
import os
import sys

import e3.fs
import e3.os.fs
import e3.sys
from e3.sys import RewriteImportNodeTransformer, RewriteImportRule, RewriteNodeError

import pytest


def test_filtering_import():
    script = """
import a, b, c
import a1, b, c
import a2, a3
from d import l1, l2, c3
from foo.bar.module import name1, name2, name3
from foo.bar2.module import name1, name2, name3
from foo.bar2.module import name2
from foo.bar2.module3 import name1
"""

    node = ast.parse(script, "<string>")
    node = RewriteImportNodeTransformer(
        [
            RewriteImportRule("b"),
            RewriteImportRule("a"),
            RewriteImportRule(".*3"),
            RewriteImportRule(r".*\.bar\..*", "name2"),
        ]
    ).visit(node)

    expected = "Module(body=["

    expected += "Import(names=[alias(name='c', asname=None)]),"
    # import a, b, c
    # b and a skipped

    expected += (
        " Import(names=[alias(name='a1', asname=None),"
        " alias(name='c', asname=None)]),"
    )
    # import a1, b, c => a1, c  -- b is skipped

    expected += " Import(names=[alias(name='a2', asname=None)]),"
    # import a2, a3 => a2 -- a3 is skipped (.*3)

    expected += (
        " ImportFrom(module='d',"
        " names=[alias(name='l1', asname=None),"
        " alias(name='l2', asname=None),"
        " alias(name='c3', asname=None)], level=0),"
    )
    # from d import l1, l2, c3 - not modified

    expected += (
        " ImportFrom(module='foo.bar.module',"
        " names=[alias(name='name1', asname=None),"
        " alias(name='name3', asname=None)], level=0),"
    )
    # from foo.bar.module import name1, name2, name3
    # .*\.bar\..* name2 -> name2 is skipped

    expected += (
        " ImportFrom(module='foo.bar2.module',"
        " names=[alias(name='name1', asname=None), "
        "alias(name='name2', asname=None), "
        "alias(name='name3', asname=None)], level=0),"
    )
    # from foo.bar2.module import name1, name2, name3 - not modifed

    expected += (
        " ImportFrom(module='foo.bar2.module',"
        " names=[alias(name='name2', asname=None)], level=0),"
    )
    # from foo.bar2.module import name2 - not modified

    expected += " ImportFrom(module='foo.bar2.module3'," " names=[], level=0)])"
    # from foo.bar2.module3 import name1 -- module matching .*3

    ast_dump = ast.dump(node)
    # Make the test compatible with Python 3.8
    ast_dump = ast_dump.replace(", type_ignores=[]", "")
    assert ast_dump == expected

    node2 = ast.parse(script, "<string>")
    with pytest.raises(RewriteNodeError) as err:
        RewriteImportNodeTransformer(
            [
                RewriteImportRule("a", action=RewriteImportRule.RuleAction.reject),
                RewriteImportRule("b"),
                RewriteImportRule(".*3"),
                RewriteImportRule(r".*\.bar\..*", "name2"),
            ]
        ).visit(node2)

    # verify that import a is rejected
    assert (
        "Import(names=[alias(name='a', asname=None),"
        " alias(name='b', asname=None),"
        " alias(name='c', asname=None)])" in str(err.value)
    )

    node3 = ast.parse(script, "<string>")
    with pytest.raises(RewriteNodeError) as err3:
        RewriteImportNodeTransformer(
            [
                RewriteImportRule("a"),
                RewriteImportRule("b"),
                RewriteImportRule(".*3"),
                RewriteImportRule(
                    r".*\.bar\..*", "name2", action=RewriteImportRule.RuleAction.reject
                ),
            ]
        ).visit(node3)
    # verify that from foo.bar.module import name2 is rejected
    assert "module='foo.bar." in str(err3.value)


def test_python_func():
    e3.sys.set_python_env("/foo")
    if sys.platform == "win32":
        assert "/foo" in os.environ["PATH"].split(os.pathsep)
        assert [e3.os.fs.unixpath(p) for p in e3.sys.python_script("run", "/foo")] == [
            "/foo/python.exe",
            "/foo/Scripts/run",
        ]
        assert [e3.os.fs.unixpath(p) for p in e3.sys.python_script("run")][
            0
        ] == e3.os.fs.unixpath(sys.executable)

        e3.fs.mkdir("Scripts")
        e3.os.fs.touch("Scripts/run.exe")
        assert e3.os.fs.unixpath(
            e3.sys.python_script("run.exe", os.getcwd())[0]
        ) == e3.os.fs.unixpath(os.path.join(os.getcwd(), "Scripts", "run.exe"))
        assert e3.os.fs.unixpath(
            e3.sys.python_script("run", os.getcwd())[0]
        ) == e3.os.fs.unixpath(os.path.join(os.getcwd(), "Scripts", "run.exe"))
    else:
        assert "/foo/bin" in os.environ["PATH"].split(os.pathsep)
        assert e3.sys.python_script("run", "/foo") == [
            "/foo/bin/python",
            "/foo/bin/run",
        ]

    assert e3.os.fs.unixpath(
        os.path.dirname(e3.sys.python_script("run")[0])
    ) == e3.os.fs.unixpath(os.path.dirname(sys.executable))
    assert e3.os.fs.unixpath(e3.sys.interpreter()) == e3.os.fs.unixpath(sys.executable)

    # Check support for python3
    if sys.platform == "win32":
        e3.os.fs.touch("python3.exe")
        assert e3.sys.interpreter(os.getcwd()) == os.path.join(
            os.getcwd(), "python3.exe"
        )
    else:
        e3.fs.mkdir("bin")
        e3.os.fs.touch(os.path.join("bin", "python3"))
        assert e3.sys.interpreter(os.getcwd()) == os.path.join(
            os.getcwd(), "bin", "python3"
        )



import os
import sys

import e3.archive
import e3.fs
import e3.log
import e3.os.fs

import pytest


@pytest.mark.parametrize("ext", (".tar.gz", ".tar.bz2", ".tar", ".zip"))
def test_unpack(ext):
    dir_to_pack = os.path.dirname(__file__)

    test_dir = os.path.basename(dir_to_pack)

    dest = "dest"
    e3.fs.mkdir(dest)

    archive_name = "e3-core" + ext

    try:
        e3.archive.create_archive(archive_name, os.path.abspath(dir_to_pack), dest)
        assert os.path.exists(os.path.join(dest, archive_name))

        with pytest.raises(e3.archive.ArchiveError):
            e3.archive.unpack_archive(
                os.path.join(dest, archive_name), os.path.join(dest, "dest")
            )

        e3.fs.mkdir(os.path.join(dest, "dest"))
        e3.archive.unpack_archive(
            os.path.join(dest, archive_name), os.path.join(dest, "dest")
        )

        assert os.path.exists(
            os.path.join(dest, "dest", test_dir, os.path.basename(__file__))
        )

        e3.fs.mkdir(os.path.join(dest, "dest2"))
        e3.archive.unpack_archive(
            os.path.join(dest, archive_name),
            os.path.join(dest, "dest2"),
            selected_files=(
                e3.os.fs.unixpath(os.path.join(test_dir, os.path.basename(__file__))),
            ),
            remove_root_dir=True,
        )

        assert os.path.exists(os.path.join(dest, "dest2", os.path.basename(__file__)))

        # Test wildcard if not .zip format
        # ??? not supported?
        if ext != ".zip":
            e3.fs.mkdir(os.path.join(dest, "dest3"))
            e3.archive.unpack_archive(
                os.path.join(dest, archive_name),
                os.path.join(dest, "dest3"),
                selected_files=(os.path.join(test_dir, "*.py"),),
                remove_root_dir=True,
            )

            assert os.path.exists(
                os.path.join(dest, "dest3", os.path.basename(__file__))
            )

        e3.archive.create_archive(
            "e3" + ext, os.path.abspath(dir_to_pack), dest, from_dir_rename="e3rename"
        )
        e3.fs.mkdir(os.path.join(dest, "dest4"))
        e3.archive.unpack_archive(
            os.path.join(dest, "e3" + ext), os.path.join(dest, "dest4")
        )
        assert os.path.join(dest, "dest4", "e3rename")

        # force use of sync_tree
        e3.fs.rm(os.path.join(dest, "dest4", "e3rename", os.path.basename(__file__)))
        e3.archive.unpack_archive(
            os.path.join(dest, "e3" + ext),
            os.path.join(dest, "dest4", "e3rename"),
            remove_root_dir=True,
        )
        assert os.path.exists(
            os.path.join(dest, "dest4", "e3rename", os.path.basename(__file__))
        )

    finally:
        e3.fs.rm(dest, True)


def test_unsupported():
    """Test unsupported archive format."""
    with pytest.raises(e3.archive.ArchiveError) as err:
        e3.archive.create_archive("foo.foo", os.getcwd(), "dest")
    assert 'unknown format "foo.foo"' in str(err)


def test_unpackerror():
    """Test unpack_archive when archive file is not found."""
    with pytest.raises(e3.archive.ArchiveError) as err:
        e3.archive.unpack_archive("foo", "dest")
    assert "cannot find foo" in str(err)


def test_unpack_cmd():
    """Test custom unpack_cmd."""
    dir_to_pack = os.path.dirname(__file__)

    dest = "dest"
    e3.fs.mkdir(dest)

    archive_name = "e3-core.tar"

    e3.archive.create_archive(archive_name, os.path.abspath(dir_to_pack), dest)

    all_dest = "all_dest"
    e3.fs.mkdir(all_dest)

    # use cp to 'extract' the archive
    e3.archive.unpack_archive(
        os.path.join(dest, archive_name), all_dest, unpack_cmd=e3.fs.cp
    )
    assert os.path.exists(os.path.join(all_dest, archive_name))

    # Use a custom unpack function and verify that it is called with
    # the expected arguments
    class TestResult:
        def store_result(self, **kwargs):
            self.kwargs = kwargs

    t = TestResult()

    def custom_unpack(filename, dest, selected_files):
        t.store_result(f=filename, d=dest, s=selected_files)

    e3.archive.unpack_archive(
        os.path.join(dest, archive_name),
        all_dest,
        unpack_cmd=custom_unpack,
        selected_files=["bar"],
    )
    assert os.path.basename(t.kwargs["f"]) == archive_name
    assert t.kwargs["d"] == all_dest
    assert t.kwargs["s"] == ["bar"]


def test_unpack_files():
    """Test unpack_archive with selected_files."""
    e3.fs.mkdir("d")
    e3.fs.mkdir(os.path.join("d/a"))
    e3.fs.mkdir(os.path.join("d/a/c"))
    e3.fs.mkdir(os.path.join("d/a/d"))
    e3.fs.mkdir(os.path.join("d/b"))

    dest = "dest"
    e3.fs.mkdir(dest)

    result_dir = "result"
    e3.fs.mkdir(result_dir)

    archive_name = "e3-core.tar"
    e3.archive.create_archive(archive_name, os.path.abspath(os.path.join("d")), dest)

    # No file starting with a path starting with 'a'.
    # Should raise ArchiveError
    with pytest.raises(e3.archive.ArchiveError):
        e3.archive.unpack_archive(
            os.path.join(dest, archive_name), result_dir, selected_files=["a"]
        )

    # unpacking d/a should work
    e3.archive.unpack_archive(
        os.path.join(dest, archive_name), result_dir, selected_files=["d/a"]
    )
    assert os.path.exists(os.path.join(result_dir, "d", "a", "c"))
    assert os.path.exists(os.path.join(result_dir, "d", "a", "d"))
    assert not os.path.exists(os.path.join(result_dir, "d", "b"))


def test_unpack_error():
    """Test unpack errors."""
    e3.fs.mkdir("dest")
    # create an invalid tgz
    e3.os.fs.touch("foo.tgz")
    with pytest.raises(e3.archive.ArchiveError) as err:
        e3.archive.unpack_archive("foo.tgz", "dest")
    assert "unpack_archive: Cannot untar" in str(err.value)

    # create an invalid zip
    e3.os.fs.touch("foo.zip")
    with pytest.raises(e3.archive.ArchiveError) as err:
        e3.archive.unpack_archive("foo.zip", "dest")
    assert "unpack_archive: Cannot unzip" in str(err.value)


def test_zip_no_root_dir():
    """Create a zip with no_root_dir."""
    e3.fs.mkdir("from")
    e3.os.fs.touch(os.path.join("from", "afile"))
    e3.fs.mkdir("dest")
    e3.archive.create_archive(
        "pkg.zip", os.path.abspath("from"), "dest", no_root_dir=True
    )
    e3.fs.mkdir("result")
    e3.archive.unpack_archive(os.path.join("dest", "pkg.zip"), "result")
    assert os.path.exists(os.path.join("result", "afile"))


def test_remove_root_dir():
    """Try create_archive no_root_dir and unpack_archive remove_root_dir."""
    e3.fs.mkdir("from")
    e3.os.fs.touch(os.path.join("from", "a"))
    e3.os.fs.touch(os.path.join("from", "b"))
    e3.fs.mkdir("dest")

    # Create an archive with two root dirs (from and dest)
    e3.archive.create_archive(
        "pkg.zip", os.path.abspath(os.getcwd()), "dest", no_root_dir=True
    )
    e3.fs.mkdir("result")

    # unpacking the archive with remote_root_dir should fail
    with pytest.raises(e3.archive.ArchiveError) as err:
        e3.archive.unpack_archive(
            os.path.join("dest", "pkg.zip"), "result", remove_root_dir=True
        )
    assert "does not have a unique root dir" in str(err)

    # In 'auto' mode, the fallback will extract the two root dirs
    e3.archive.unpack_archive(
        os.path.join("dest", "pkg.zip"), "result", remove_root_dir="auto"
    )
    assert os.path.exists(os.path.join("result", "from", "a"))
    assert os.path.exists(os.path.join("result", "dest"))

    # Running it again with use sync_tree
    e3.fs.rm(os.path.join("result", "from", "a"))
    e3.archive.unpack_archive(
        os.path.join("dest", "pkg.zip"), "result", remove_root_dir="auto"
    )
    assert os.path.exists(os.path.join("result", "from", "a"))


def test_empty():
    """Create an archive with an empty content."""
    e3.fs.mkdir("from")
    e3.fs.mkdir("dest")
    e3.fs.mkdir("result")
    e3.archive.create_archive(
        "pkg.zip", os.path.abspath("from"), "dest", no_root_dir=True
    )

    # Remove root dir should be a noop
    e3.archive.unpack_archive(
        os.path.join("dest", "pkg.zip"), "result", remove_root_dir=True
    )
    assert os.listdir("result") == []


@pytest.mark.skipif(sys.platform == "win32", reason="test executable attribute")
def test_zip_attributes():
    zip_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test.zip")
    e3.fs.mkdir("result")
    e3.archive.unpack_archive(zip_file, "result", remove_root_dir=True)
    assert os.access("result/test.sh", os.X_OK)


def test_archive_with_readonly_dir():
    """Test unpack of archive with read-only directory."""
    e3.fs.mkdir("from")
    e3.fs.mkdir("dest")
    e3.fs.mkdir("result")
    e3.fs.mkdir("from/readonly_dir")
    e3.os.fs.touch("from/readonly_dir/file.txt")
    e3.os.fs.chmod("u=rx,go=rx", "from/readonly_dir")
    e3.archive.create_archive("pkg.tar.gz", os.path.abspath("from"), "dest")
    e3.os.fs.chmod("urwx", "from/readonly_dir")
    e3.archive.unpack_archive(
        os.path.join("dest", "pkg.tar.gz"), "result", remove_root_dir=True
    )
    assert os.path.isfile(os.path.join("result", "readonly_dir", "file.txt"))



from __future__ import annotations

import heapq
import time
import typing
from datetime import datetime
from queue import Empty, Queue
from typing import TYPE_CHECKING

import e3.log
from e3.collection.dag import DAGIterator
from e3.job import Job


if TYPE_CHECKING:
    from typing import Any, Callable, Dict, FrozenSet, List, Optional, Tuple, Type
    from e3.collection.dag import DAG

    JobProviderCallback = Callable[
        [str, Any, FrozenSet[str], Callable[[str], None]], Job
    ]
    CollectCallback = Callable[[Job], bool]


logger = e3.log.getLogger("job.scheduler")

# The default maximum duration for a job, in seconds (24 hours).
DEFAULT_JOB_MAX_DURATION = 3600 * 24


class Scheduler:
    """Handle parallel execution of interdependent jobs."""

    def __init__(
        self,
        job_provider: JobProviderCallback,
        collect: Optional[CollectCallback] = None,
        queues: Optional[Dict[str, int]] = None,
        tokens: int = 1,
        job_timeout: int = DEFAULT_JOB_MAX_DURATION,
    ):
        """Initialize Scheduler.

        :param job_provider: function that returns instances of Job.
            The function takes as arguments: the job uid, the data
            associated with it, the list of predecessors id and and
            a notification function called when the job end.
        :param collect: function that collect results from the
            jobs. If the function returns True then the job is requeued
        :param queues: describes the list of queues handled by the
            scheduler. The format is a dictionary for which which
            keys are queue names and value the max number of tokens
            available at a given time. if None then a single queue
            called "default" is created. Its size is then given by
            the tokens parameter.
        :param tokens: number of tokens for the default queue. Relevant
            only when queues is None
        :param job_timeout: maximum execution time for a job. The default
            is 24h. If set to None timeout are disabled but it also make
            the scheduller non interruptable when waiting for a job to
            finish.
        """
        self.job_provider = job_provider
        self.job_timeout = job_timeout
        if collect is None:
            self.collect: Callable[[Job], bool] = lambda x: False
        else:
            self.collect = collect

        self.active_jobs: List[Job] = []
        self.queued_jobs = 0
        self.all_jobs_queued = False
        self.message_queue: Queue[Any] = Queue()
        self.dag: Optional[DAG] = None
        self.dag_iterator: Optional[DAGIterator] = None
        self.start_time: Optional[datetime] = None
        self.stop_time: Optional[datetime] = None
        self.max_active_jobs = 0

        # Initialize named queues
        self.queues: Dict[str, List[Tuple[int, int, Job]]] = {}
        self.tokens = {}
        self.n_tokens = 0

        if queues is None:
            # If no queues are specificied create a default one
            # with no name.
            queues = {"default": tokens}

        self.global_queue_index = 1

        # Create the queues
        for name, max_token in queues.items():
            self.queues[name] = []
            self.tokens[name] = max_token
            self.n_tokens += max_token

        # Create a slot reserve. The goal is to give a given job a number
        # which is unique among the active jobs. We need a maximum of
        # self.n_tokens slots
        self.slots = list(range(self.n_tokens))

    def safe_collect(self, job: Job) -> bool:
        """Protect call to collect.

        This ensures for job such as JobProcess that there are no calls to
        Run during collect. Main goal is to avoid leak of handles from collect
        to a process spawned by a Job. On Unixes consequences of such leak is
        more a security concern than an operational one. On Windows, this can
        lead easily to file locking issues and thus might cause crashes.
        """
        with Job.lock:
            return self.collect(job)

    def safe_job_provider(
        self,
        uid: str,
        data: Any,
        predecessors: FrozenSet[str],
        notify_end: Callable[[str], None],
    ) -> Job:
        """Protect call to job_provider.

        See safe_collect commment above.
        """
        with Job.lock:
            return self.job_provider(uid, data, predecessors, notify_end)

    @classmethod
    def simple_provider(cls, job_class: Type[Job]) -> JobProviderCallback:
        """Return a simple provider based on a given Job class.

        :param job_class: a subclass of Job
        """

        def provider(
            uid: str,
            data: Any,
            predecessors: FrozenSet[str],
            notify_end: Callable[[str], None],
        ) -> Job:
            del predecessors
            return job_class(uid, data, notify_end)

        return provider

    def init_state(self, dag: DAG) -> None:
        """Reinitialize the scheduler state (internal function).

        :param dag: the dag representing the list of job to execute
        """
        # Active jobs
        self.active_jobs = []

        # Total number of jobs in the queues
        self.queued_jobs = 0

        # Have all jobs been queued?
        self.all_jobs_queued = False

        # Message queue to get job end notifications
        self.message_queue = Queue()

        self.dag = dag
        self.dag_iterator = DAGIterator(dag, enable_busy_state=True)
        self.start_time = datetime.now()
        self.stop_time = None
        self.max_active_jobs = 0

    @property
    def is_finished(self) -> bool:
        """Check if all jobs have been executed (internal).

        :return: True if complete
        """
        # The run is considered finished once there is no more job
        # in the DAG, the queues and that no job is running.
        return self.all_jobs_queued and self.queued_jobs == 0 and not self.active_jobs

    def log_state(self) -> None:
        """Log the current state of the scheduler (internal)."""
        logger.debug(
            "non-ready?: %s, in queue: %s, running: %s",
            not self.all_jobs_queued,
            self.queued_jobs,
            len(self.active_jobs),
        )

    def run(self, dag: DAG) -> None:
        """Launch the scheduler."""
        self.init_state(dag)

        try:
            while not self.is_finished:
                self.enqueue()
                self.launch()
                self.log_state()
                self.max_active_jobs = max(self.max_active_jobs, len(self.active_jobs))
                self.wait()
        except KeyboardInterrupt:
            logger.info("Interrupting jobs...")
            for p in self.active_jobs:
                p.interrupt()
                self.safe_collect(p)
                p.on_finish(self)
            raise KeyboardInterrupt
        self.stop_time = datetime.now()

    def push(self, job: Job) -> None:
        """Push a job into a queue."""
        # We use a tuple here to ensure the stability of the queue
        # when two jobs have similar priorities.
        heapq.heappush(
            self.queues[job.queue_name], (-job.priority, self.global_queue_index, job)
        )
        self.global_queue_index += 1
        self.queued_jobs += 1

    def enqueue(self) -> None:
        """Push into the queues job that are ready (internal)."""
        if self.all_jobs_queued:
            return

        if TYPE_CHECKING:
            assert self.dag_iterator is not None

        try:
            while True:
                uid, data, predecessors = self.dag_iterator.next_element()
                if uid is None:
                    # No more jobs ready
                    return

                job = self.safe_job_provider(
                    typing.cast(str, uid),
                    data,
                    predecessors=typing.cast(typing.FrozenSet[str], predecessors),
                    notify_end=self.message_queue.put,
                )
                if job.should_skip:
                    self.safe_collect(job)
                    job.on_finish(self)
                    self.dag_iterator.leave(uid)
                else:
                    self.push(job)
        except StopIteration:
            self.all_jobs_queued = True

    def launch(self) -> None:
        """Launch next jobs in the queues (internal)."""
        if self.queued_jobs == 0:
            return

        for name in self.queues:
            q = self.queues[name]
            while q and q[0][2].tokens <= self.tokens[name]:
                _, _, next_job = heapq.heappop(q)
                self.queued_jobs -= 1
                next_job.on_start(self)
                next_job.start(slot=self.slots.pop())
                self.tokens[name] -= next_job.tokens
                self.active_jobs.append(next_job)

    def wait(self) -> None:
        """Wait for the end of an active job."""
        if not self.active_jobs:
            return

        if TYPE_CHECKING:
            assert self.dag is not None
            assert self.dag_iterator is not None

        # Wait for message from one the active jobs
        while True:
            # The first job in active jobs is the oldest one
            # compute the get timeout based on its startup information
            first_job = self.active_jobs[0]
            current_timeout = self.job_timeout - first_job.timing_info.duration

            # Ensure waiting time is a positive number
            current_timeout = max(0.1, current_timeout)

            try:
                uid = self.message_queue.get(block=True, timeout=current_timeout)

            except Empty:
                # If after timeout we get an empty result, it means that
                # the oldest job has reached the timeout. Interrupt it
                # and wait for the queue to receive the end notification
                self.active_jobs[0].interrupt()
                time.sleep(0.1)

            else:
                job_index, job = next(
                    (
                        (index, job)
                        for index, job in enumerate(self.active_jobs)
                        if job.uid == uid
                    )
                )
                ti = job.timing_info
                logger.debug(
                    "job %s %s after %s",
                    uid,
                    "interrupted" if job.interrupted else "finished",
                    ti.duration,
                )
                self.slots.append(job.slot)

                # Liberate the resources taken by the job
                self.tokens[job.queue_name] += job.tokens
                collect_result = self.safe_collect(job)
                job.on_finish(self)

                if collect_result:
                    # Requeue when needed
                    self.push(job)
                else:
                    # Mark the job as completed
                    if job.uid in self.dag:
                        self.dag_iterator.leave(job.uid)

                del self.active_jobs[job_index]
                return



import sys

from e3.collection.dag import DAG
from e3.job import Job, ProcessJob
from e3.job.scheduler import Scheduler

import pytest


class NopJob(Job):
    def run(self):
        pass

    @property
    def priority(self):
        try:
            result = -int(self.uid)
        except Exception:
            result = 0
        return result


class SleepJob(ProcessJob):
    @property
    def cmdline(self):
        return [sys.executable, "-c", "import time; time.sleep(6.0)"]


class TestScheduler:
    def test_minimal_run(self):
        """Test with only two independent jobs."""
        dag = DAG()
        dag.add_vertex("1")
        dag.add_vertex("2")
        s = Scheduler(Scheduler.simple_provider(NopJob), tokens=2)
        s.run(dag)
        assert s.max_active_jobs == 2

    def test_ordering(self):
        """Test that jobs are ordered correctly."""
        results = []

        def collect(job):
            results.append(job.uid)

        dag = DAG()
        dag.add_vertex("3")
        dag.add_vertex("0")
        dag.add_vertex("1")
        s = Scheduler(Scheduler.simple_provider(NopJob), tokens=1, collect=collect)
        s.run(dag)
        assert tuple(results) == ("0", "1", "3")

    def test_minimal_run2(self):
        """Test with two interdependent jobs."""
        dag = DAG()
        dag.add_vertex("1")
        dag.add_vertex("2", predecessors=["1"])
        s = Scheduler(Scheduler.simple_provider(NopJob), tokens=2)
        s.run(dag)
        assert s.max_active_jobs == 1

    def test_requeue(self):
        """Requeue test.

        Same as previous example except that all tests are requeued
        once.
        """
        results = {}

        def collect(job):
            if job.uid not in results:
                results[job.uid] = True
                return True
            else:
                return False

        # This time test with two interdependent jobs
        dag = DAG()
        dag.add_vertex("1")
        dag.add_vertex("2")
        s = Scheduler(Scheduler.simple_provider(NopJob), tokens=2, collect=collect)
        s.run(dag)
        assert s.max_active_jobs == 2
        assert results["1"]
        assert results["2"]

    def test_skip(self):
        """Simple example in which all the tests are skipped."""
        results = {}

        def get_job(uid, data, predecessors, notify_end):
            result = NopJob(uid, data, notify_end)
            result.should_skip = True
            return result

        def collect(job):
            results[job.uid] = job.timing_info

        # This time test with two interdependent jobs
        dag = DAG()
        dag.add_vertex("1")
        dag.add_vertex("2")
        s = Scheduler(get_job, tokens=2, collect=collect)
        s.run(dag)

        # Check start_time end_time to be sure tests have not been run
        for v in results.values():
            assert v.start_time is None
            assert v.stop_time is None

    def test_timeout(self):
        """Ensure that jobs are interrupted correctly on timeout."""
        results = {}
        pytest.importorskip("psutil")

        def get_job(uid, data, predecessors, notify_end):
            return SleepJob(uid, data, notify_end)

        def collect(job):
            results[job.uid] = job

        dag = DAG()
        dag.add_vertex("1")
        dag.add_vertex("2")
        s = Scheduler(get_job, tokens=2, collect=collect, job_timeout=2)
        s.run(dag)

        for v in results.values():
            assert v.interrupted

    def test_keyboard_interrupt(self):
        """Ensure that jobs can be interrupted."""
        results = {}
        pytest.importorskip("psutil")

        def get_job(uid, data, predecessors, notify_end):
            return NopJob(uid, data, notify_end)

        def collect(job):
            results[job.uid] = job

        dag = DAG()
        dag.add_vertex("1")
        dag.add_vertex("2")
        s = Scheduler(get_job, tokens=2, collect=collect, job_timeout=2)

        # fake log_state that will raise a KeyboardInterrupt
        def fake_log_state():
            raise KeyboardInterrupt

        s.log_state = fake_log_state

        with pytest.raises(KeyboardInterrupt):
            s.run(dag)

        for v in results.values():
            assert v.interrupted

    def test_collect_feedback_scheme(self):
        """Collect feedback construction.

        Scheme in which if a job predecessor "fails" then job is skipped
        In order to do that get_job and collect should have access to
        common data. Note that scheduler ensure that these functions
        are called sequentially.
        """

        class SchedulerContext:
            def __init__(self):
                # Save in results tuples with first element being a bool
                # indicating success or failure and the second the job itself
                self.results = {}

            def get_job(self, uid, data, predecessors, notify_end):
                result = NopJob(uid, data, notify_end)

                # If any of the predecessor failed skip the job
                for k in predecessors:
                    if not self.results[k][0]:
                        result.should_skip = True
                return result

            def collect(self, job):
                if job.should_skip:
                    # Skipped jobs are considered failed
                    self.results[job.uid] = [False, job]
                else:
                    # Job '2' is always failing
                    if job.uid == "2":
                        self.results[job.uid] = [False, job]
                    else:
                        self.results[job.uid] = [True, job]

        dag = DAG()
        dag.add_vertex("1")
        dag.add_vertex("2")
        dag.add_vertex("3", predecessors=["1", "2"])
        dag.add_vertex("4", predecessors=["3"])
        c = SchedulerContext()
        s = Scheduler(c.get_job, tokens=2, collect=c.collect)
        s.run(dag)

        assert (
            not c.results["2"][1].should_skip and not c.results["2"][0]
        ), 'job "2" is run and should be marked as failed'
        assert c.results["3"][1].should_skip, 'job "3" should be skipped'
        assert c.results["4"][1].should_skip, 'job "4" should be skipped'



#!/usr/bin/env python
# type: ignore
"""Generate the requirement coverage for e3-core doc."""

import sys
import yaml


def lookup(item, d):
    """Lookup an item in a dictionary."""
    return {key for key, value in d.items() if value == item}


def merge_docs(requirement: str, coverage: str) -> dict:
    """Load requirement yaml and include the coverage info.

    :param requirement: yaml filename containing list of requirements
    :param coverage: yaml filename formatting as 'testname: requirement'
    """
    with open(requirement) as f:
        reqs = yaml.safe_load(f)

    with open(coverage) as f:
        reqs_cov = yaml.safe_load(f)

    for k in reqs:
        tests = lookup(k, reqs_cov)
        reqs[k]["tests"] = tests
    return reqs


def generate_rst(reqs_result: dict, dest: str):
    """Generate rst file from requirement coverage data.

    :param reqs_result: dictionary returned by merge_docs
    :param dest: rst file to create
    """
    with open(dest, "w") as f:
        for k in reqs_result:
            f.write("- %s\n" % k)
            f.write("  %s\n" % reqs_result[k]["desc"].encode("utf-8"))
            tests = reqs_result[k]["tests"]
            if tests:
                f.write("  **Covered by %s**\n" % ", ".join(reqs_result[k]["tests"]))
            else:
                f.write("  **Not yet covered**\n")


if __name__ == "__main__":
    generate_rst(merge_docs(sys.argv[1], sys.argv[2]), sys.argv[3])
